from __future__ import annotations

import ast
import os
import stat
import threading
from collections import defaultdict
from dataclasses import dataclass, field
from functools import lru_cache
from pathlib import Path
from typing import Generator

from tach import errors
from tach import filesystem as fs
from tach.colors import BCOLORS
from tach.constants import ROOT_MODULE_SENTINEL_TAG
from tach.core import ModuleConfig


@dataclass
class FileInfo:
    path: str
    content: str | None = None
    canonical_path: str | None = None
    ast: ast.AST | None = None


# Thread-local file cache to avoid going to disk as much as possible
thread_local = threading.local()
# Cannot type-hint non-self attributes (https://github.com/python/mypy/issues/2388)
# cwd: str
thread_local.cwd = os.getcwd()
# file_caches_by_cwd: defaultdict[str, dict[str, FileInfo]]
thread_local.file_caches_by_cwd = defaultdict(dict)


def get_cwd() -> str:
    # Use a cached cwd to avoid system calls
    if not hasattr(thread_local, "cwd"):
        thread_local.cwd = os.getcwd()
    return thread_local.cwd


def chdir(path: str):
    # When using chdir, update the cached version
    os.chdir(path)
    thread_local.cwd = os.getcwd()


def _get_file_cache() -> dict[str, FileInfo]:
    if not hasattr(thread_local, "file_caches_by_cwd"):
        thread_local.file_caches_by_cwd = defaultdict(dict)
    file_caches_by_cwd: defaultdict[str, dict[str, FileInfo]] = (
        thread_local.file_caches_by_cwd
    )  # type: ignore
    return file_caches_by_cwd[get_cwd()]


def _file_cache_key(path: str) -> str:
    return f"{get_cwd()}:::{path}"


def _cached_file(path: str) -> FileInfo | None:
    return _get_file_cache().get(_file_cache_key(path))


def _set_cached_file(path: str, file_info: FileInfo):
    _get_file_cache()[_file_cache_key(path)] = file_info


def _remove_cached_file(path: str):
    _get_file_cache().pop(_file_cache_key(path), None)


def canonical(path: str) -> str:
    cached_file = _cached_file(path)
    if cached_file and cached_file.canonical_path:
        return cached_file.canonical_path

    result = os.path.relpath(os.path.realpath(path), start=get_cwd())

    if cached_file:
        cached_file.canonical_path = result
    else:
        _set_cached_file(path, FileInfo(path=path, canonical_path=result))

    return result


def read_file(path: str) -> str:
    cached_file = _cached_file(path)
    if cached_file and cached_file.content:
        return cached_file.content

    with open(path) as f:
        content = f.read()

    if cached_file:
        cached_file.content = content
    else:
        _set_cached_file(path, FileInfo(path=path, content=content))

    return content


def write_file(path: str, content: str):
    with open(path, "w") as f:
        f.write(content)
        print(f"{BCOLORS.WARNING}Wrote '{canonical(path)}'{BCOLORS.ENDC}")

    cached_file = _cached_file(path)
    if cached_file:
        cached_file.content = content
        cached_file.ast = None
    else:
        _set_cached_file(path, FileInfo(path=path, content=content))


def delete_file(path: str):
    _remove_cached_file(path)
    os.unlink(path)
    print(f"{BCOLORS.WARNING}Deleted '{canonical(path)}'{BCOLORS.ENDC}")


def mark_executable(path: str):
    file_path = Path(path)
    file_path.chmod(
        file_path.stat().st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    )


def parse_ast(path: str) -> ast.AST:
    cached_file = _cached_file(path)
    if cached_file and cached_file.ast:
        return cached_file.ast

    if cached_file and cached_file.content:
        content = cached_file.content
        try:
            ast_result = ast.parse(cached_file.content)
        except SyntaxError as e:
            raise errors.TachParseError(f"Syntax error in {path}: {e}")
    else:
        with open(path) as f:
            content = f.read()
        try:
            ast_result = ast.parse(content)
        except SyntaxError as e:
            raise errors.TachParseError(f"Syntax error in {path}: {e}")

    if cached_file:
        cached_file.content = content
        cached_file.ast = ast_result
    else:
        _set_cached_file(path, FileInfo(path=path, content=content, ast=ast_result))

    return ast_result


def walk(
    root: Path, depth: int | None = None
) -> Generator[tuple[Path, list[Path]], None, None]:
    if depth is not None and depth <= 0:
        return
    root = root.resolve()
    for dirpath, dirnames, filenames in os.walk(root):
        rel_dirpath = Path(dirpath).relative_to(root)

        if rel_dirpath.name.startswith("."):
            # This prevents recursing into child directories of hidden paths
            del dirnames[:]
            continue

        if depth:
            # Ignore anything past requested depth
            current_depth = len(rel_dirpath.parts) - 1
            if current_depth > depth:
                continue

        def filter_filename(filename: str) -> bool:
            return not filename.startswith(".")

        yield rel_dirpath, list(map(Path, filter(filter_filename, filenames)))


def walk_pyfiles(root: Path, depth: int | None = None) -> Generator[Path, None, None]:
    for dirpath, filepaths in walk(root, depth=depth):
        for filepath in filepaths:
            if filepath.name.endswith(".py"):
                yield dirpath / filepath


@lru_cache(maxsize=None)
def file_to_module_path(source_roots: tuple[Path, ...], file_path: Path) -> str:
    # NOTE: source_roots are assumed to be absolute here
    matching_root: Path | None = None
    for root in source_roots:
        if root in file_path.parents:
            matching_root = root
            break

    if matching_root is None:
        raise ValueError(f"File path: {file_path} not found in any source root.")

    relative_path = file_path.relative_to(matching_root)
    components = list(relative_path.parent.parts)

    if relative_path.name != "__init__.py":
        components.append(relative_path.stem)

    module_path = ".".join(components)
    return "." if not module_path else module_path


@lru_cache(maxsize=None)
def module_to_interface_path(
    source_roots: tuple[Path, ...], module_path: str
) -> Path | None:
    """
    This resolves a dotted Python module path ('a.b.c')
    into a Python file path or a Python package __init__.py
    which could contain the public interface of the module.

    The module path is assumed NOT to refer to a member within a Python module.

    'source_roots' is assumed to be a list of absolute paths.
    """
    if module_path == ROOT_MODULE_SENTINEL_TAG:
        return None

    base_path = module_path.replace(".", os.sep)
    for source_root in source_roots:
        init_pyi_path = source_root / base_path / "__init__.pyi"
        init_py_path = source_root / base_path / "__init__.py"
        pyinterface_path = source_root / f"{base_path}.pyi"
        pyfile_path = source_root / f"{base_path}.py"

        if init_pyi_path.exists():
            return init_pyi_path
        elif init_py_path.exists():
            return init_py_path
        elif pyinterface_path.exists():
            return pyinterface_path
        elif pyfile_path.exists():
            return pyfile_path

    return None


@lru_cache(maxsize=None)
def module_to_pyfile_or_dir_path(
    source_roots: tuple[Path, ...], module_path: str
) -> Path | None:
    """
    This resolves a dotted Python module path ('a.b.c')
    into a Python file or a Python package directory,
    used in cases where __init__.py is not relevant such as the
    interactive module tree.

    The module path is assumed NOT to refer to a member within a Python module.

    'source_roots' is assumed to be a list of absolute paths.
    """
    if not module_path:
        # Path("") turns into PosixPath("."), but we don't want to
        # treat an empty module path as the root directory
        return None

    base_path = module_path.replace(".", os.sep)
    for source_root in source_roots:
        dir_path = source_root / base_path
        pyinterface_path = source_root / f"{base_path}.pyi"
        pyfile_path = source_root / f"{base_path}.py"
        if dir_path.is_dir():
            return dir_path
        elif pyinterface_path.exists():
            return pyinterface_path
        elif pyfile_path.exists():
            return pyfile_path

    return None


@dataclass
class ProjectModuleValidationResult:
    valid_modules: list[ModuleConfig] = field(default_factory=list)
    invalid_modules: list[ModuleConfig] = field(default_factory=list)


def validate_project_modules(
    source_roots: list[Path],
    modules: list[ModuleConfig],
) -> ProjectModuleValidationResult:
    result = ProjectModuleValidationResult()
    for module in modules:
        # NOTE: should probably also (optionally?) validate that no source root contains __init__.py
        if module.path == ROOT_MODULE_SENTINEL_TAG or fs.module_to_pyfile_or_dir_path(
            source_roots=tuple(source_roots), module_path=module.path
        ):
            result.valid_modules.append(module)
        else:
            result.invalid_modules.append(module)
    return result



from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

import pytest

from tach.core.config import Dependency, ModuleConfig, ProjectConfig, RootModuleConfig
from tach.core.modules import ModuleNode, ModuleTree
from tach.test import get_affected_modules, get_changed_module_paths

# Module-related data in fixtures was auto-generated from the 'tach' codebase
# June 28, 2024


@pytest.fixture
def modules() -> list[ModuleConfig]:
    return [
        ModuleConfig(path="tach", depends_on=[], strict=True),
        ModuleConfig(
            path="tach.__main__",
            depends_on=[Dependency(path="tach.start")],
            strict=True,
        ),
        ModuleConfig(
            path="tach.cache",
            depends_on=[Dependency(path="tach"), Dependency(path="tach.filesystem")],
            strict=True,
        ),
        ModuleConfig(
            path="tach.check",
            depends_on=[
                Dependency(path="tach.errors"),
                Dependency(path="tach.filesystem"),
                Dependency(path="tach.parsing"),
            ],
            strict=True,
        ),
        ModuleConfig(
            path="tach.cli",
            depends_on=[
                Dependency(path="tach"),
                Dependency(path="tach.cache"),
                Dependency(path="tach.check"),
                Dependency(path="tach.colors"),
                Dependency(path="tach.constants"),
                Dependency(path="tach.core"),
                Dependency(path="tach.errors"),
                Dependency(path="tach.filesystem"),
                Dependency(path="tach.logging"),
                Dependency(path="tach.mod"),
                Dependency(path="tach.parsing"),
                Dependency(path="tach.report"),
                Dependency(path="tach.show"),
                Dependency(path="tach.sync"),
                Dependency(path="tach.test"),
            ],
            strict=True,
        ),
        ModuleConfig(path="tach.colors", depends_on=[], strict=True),
        ModuleConfig(path="tach.constants", depends_on=[], strict=True),
        ModuleConfig(
            path="tach.core",
            depends_on=[Dependency(path="tach.constants")],
            strict=True,
        ),
        ModuleConfig(path="tach.errors", depends_on=[], strict=True),
        ModuleConfig(
            path="tach.filesystem",
            depends_on=[
                Dependency(path="tach.colors"),
                Dependency(path="tach.constants"),
                Dependency(path="tach.core"),
                Dependency(path="tach.errors"),
                Dependency(path="tach.hooks"),
            ],
            strict=True,
        ),
        ModuleConfig(
            path="tach.filesystem.git_ops",
            depends_on=[Dependency(path="tach.errors")],
            strict=True,
        ),
        ModuleConfig(
            path="tach.hooks",
            depends_on=[Dependency(path="tach.constants")],
            strict=True,
        ),
        ModuleConfig(
            path="tach.interactive",
            depends_on=[
                Dependency(path="tach.errors"),
                Dependency(path="tach.filesystem"),
            ],
            strict=True,
        ),
        ModuleConfig(
            path="tach.logging",
            depends_on=[
                Dependency(path="tach"),
                Dependency(path="tach.cache"),
                Dependency(path="tach.parsing"),
            ],
            strict=True,
        ),
        ModuleConfig(
            path="tach.mod",
            depends_on=[
                Dependency(path="tach.colors"),
                Dependency(path="tach.constants"),
                Dependency(path="tach.errors"),
                Dependency(path="tach.filesystem"),
                Dependency(path="tach.interactive"),
                Dependency(path="tach.parsing"),
            ],
            strict=True,
        ),
        ModuleConfig(
            path="tach.parsing",
            depends_on=[
                Dependency(path="tach.constants"),
                Dependency(path="tach.core"),
                Dependency(path="tach.filesystem"),
            ],
            strict=True,
        ),
        ModuleConfig(
            path="tach.report", depends_on=[Dependency(path="tach.errors")], strict=True
        ),
        ModuleConfig(path="tach.show", depends_on=[], strict=True),
        ModuleConfig(
            path="tach.start", depends_on=[Dependency(path="tach.cli")], strict=True
        ),
        ModuleConfig(
            path="tach.sync",
            depends_on=[
                Dependency(path="tach.check"),
                Dependency(path="tach.errors"),
                Dependency(path="tach.filesystem"),
                Dependency(path="tach.parsing"),
            ],
            strict=True,
        ),
        ModuleConfig(
            path="tach.test",
            depends_on=[
                Dependency(path="tach.errors"),
                Dependency(path="tach.filesystem"),
                Dependency(path="tach.filesystem.git_ops"),
                Dependency(path="tach.parsing"),
            ],
            strict=False,
        ),
    ]


@pytest.fixture
def module_tree() -> ModuleTree:
    return ModuleTree(
        root=ModuleNode(
            is_end_of_path=True,
            full_path=".",
            config=RootModuleConfig(path="<root>", depends_on=[], strict=False),
            interface_members=[],
            children={
                "tach": ModuleNode(
                    is_end_of_path=True,
                    full_path="tach",
                    config=ModuleConfig(path="tach", depends_on=[], strict=True),
                    interface_members=["__version__"],
                    children={
                        "__main__": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.__main__",
                            config=ModuleConfig(
                                path="tach.__main__",
                                depends_on=[Dependency(path="tach.start")],
                                strict=True,
                            ),
                            interface_members=[],
                            children={},
                        ),
                        "cache": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.cache",
                            config=ModuleConfig(
                                path="tach.cache",
                                depends_on=[
                                    Dependency(path="tach"),
                                    Dependency(path="tach.filesystem"),
                                ],
                                strict=True,
                            ),
                            interface_members=[
                                "get_uid",
                                "update_latest_version",
                                "get_latest_version",
                            ],
                            children={},
                        ),
                        "check": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.check",
                            config=ModuleConfig(
                                path="tach.check",
                                depends_on=[
                                    Dependency(path="tach.errors"),
                                    Dependency(path="tach.filesystem"),
                                    Dependency(path="tach.parsing"),
                                ],
                                strict=True,
                            ),
                            interface_members=["BoundaryError", "check"],
                            children={},
                        ),
                        "cli": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.cli",
                            config=ModuleConfig(
                                path="tach.cli",
                                depends_on=[
                                    Dependency(path="tach"),
                                    Dependency(path="tach.cache"),
                                    Dependency(path="tach.check"),
                                    Dependency(path="tach.colors"),
                                    Dependency(path="tach.constants"),
                                    Dependency(path="tach.core"),
                                    Dependency(path="tach.errors"),
                                    Dependency(path="tach.filesystem"),
                                    Dependency(path="tach.logging"),
                                    Dependency(path="tach.mod"),
                                    Dependency(path="tach.parsing"),
                                    Dependency(path="tach.report"),
                                    Dependency(path="tach.show"),
                                    Dependency(path="tach.sync"),
                                    Dependency(path="tach.test"),
                                ],
                                strict=True,
                            ),
                            interface_members=["main"],
                            children={},
                        ),
                        "colors": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.colors",
                            config=ModuleConfig(
                                path="tach.colors", depends_on=[], strict=True
                            ),
                            interface_members=["BCOLORS"],
                            children={},
                        ),
                        "constants": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.constants",
                            config=ModuleConfig(
                                path="tach.constants", depends_on=[], strict=True
                            ),
                            interface_members=[
                                "PACKAGE_NAME",
                                "TOOL_NAME",
                                "CONFIG_FILE_NAME",
                                "PACKAGE_FILE_NAME",
                                "ROOT_MODULE_SENTINEL_TAG",
                                "DEFAULT_EXCLUDE_PATHS",
                            ],
                            children={},
                        ),
                        "core": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.core",
                            config=ModuleConfig(
                                path="tach.core",
                                depends_on=[Dependency(path="tach.constants")],
                                strict=True,
                            ),
                            interface_members=[
                                "ProjectConfig",
                                "ModuleConfig",
                                "ModuleNode",
                                "ModuleTree",
                                "UnusedDependencies",
                            ],
                            children={},
                        ),
                        "errors": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.errors",
                            config=ModuleConfig(
                                path="tach.errors", depends_on=[], strict=True
                            ),
                            interface_members=[
                                "TachError",
                                "TachParseError",
                                "TachSetupError",
                            ],
                            children={},
                        ),
                        "filesystem": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.filesystem",
                            config=ModuleConfig(
                                path="tach.filesystem",
                                depends_on=[
                                    Dependency(path="tach.colors"),
                                    Dependency(path="tach.constants"),
                                    Dependency(path="tach.core"),
                                    Dependency(path="tach.errors"),
                                    Dependency(path="tach.hooks"),
                                ],
                                strict=True,
                            ),
                            interface_members=[
                                "get_cwd",
                                "chdir",
                                "canonical",
                                "read_file",
                                "write_file",
                                "delete_file",
                                "parse_ast",
                                "walk",
                                "walk_pyfiles",
                                "file_to_module_path",
                                "module_to_file_path_no_members",
                                "module_to_pyfile_or_dir_path",
                                "get_project_config_path",
                                "find_project_config_root",
                                "install_pre_commit",
                                "validate_project_modules",
                                "ProjectModuleValidationResult",
                            ],
                            children={
                                "git_ops": ModuleNode(
                                    is_end_of_path=True,
                                    full_path="tach.filesystem.git_ops",
                                    config=ModuleConfig(
                                        path="tach.filesystem.git_ops",
                                        depends_on=[Dependency(path="tach.errors")],
                                        strict=True,
                                    ),
                                    interface_members=["get_changed_files"],
                                    children={},
                                )
                            },
                        ),
                        "hooks": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.hooks",
                            config=ModuleConfig(
                                path="tach.hooks",
                                depends_on=[Dependency(path="tach.constants")],
                                strict=True,
                            ),
                            interface_members=["build_pre_commit_hook_content"],
                            children={},
                        ),
                        "interactive": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.interactive",
                            config=ModuleConfig(
                                path="tach.interactive",
                                depends_on=[
                                    Dependency(path="tach.errors"),
                                    Dependency(path="tach.filesystem"),
                                ],
                                strict=True,
                            ),
                            interface_members=[
                                "get_selected_modules_interactive",
                                "InteractiveModuleConfiguration",
                            ],
                            children={},
                        ),
                        "logging": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.logging",
                            config=ModuleConfig(
                                path="tach.logging",
                                depends_on=[
                                    Dependency(path="tach"),
                                    Dependency(path="tach.cache"),
                                    Dependency(path="tach.parsing"),
                                ],
                                strict=True,
                            ),
                            interface_members=["logger", "LogDataModel"],
                            children={},
                        ),
                        "mod": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.mod",
                            config=ModuleConfig(
                                path="tach.mod",
                                depends_on=[
                                    Dependency(path="tach.colors"),
                                    Dependency(path="tach.constants"),
                                    Dependency(path="tach.errors"),
                                    Dependency(path="tach.filesystem"),
                                    Dependency(path="tach.interactive"),
                                    Dependency(path="tach.parsing"),
                                ],
                                strict=True,
                            ),
                            interface_members=["mod_edit_interactive"],
                            children={},
                        ),
                        "parsing": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.parsing",
                            config=ModuleConfig(
                                path="tach.parsing",
                                depends_on=[
                                    Dependency(path="tach.constants"),
                                    Dependency(path="tach.core"),
                                    Dependency(path="tach.filesystem"),
                                ],
                                strict=True,
                            ),
                            interface_members=[
                                "parse_project_config",
                                "dump_project_config_to_yaml",
                                "parse_interface_members",
                                "build_module_tree",
                            ],
                            children={},
                        ),
                        "report": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.report",
                            config=ModuleConfig(
                                path="tach.report",
                                depends_on=[Dependency(path="tach.errors")],
                                strict=True,
                            ),
                            interface_members=["report"],
                            children={},
                        ),
                        "show": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.show",
                            config=ModuleConfig(
                                path="tach.show", depends_on=[], strict=True
                            ),
                            interface_members=["generate_show_url"],
                            children={},
                        ),
                        "start": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.start",
                            config=ModuleConfig(
                                path="tach.start",
                                depends_on=[Dependency(path="tach.cli")],
                                strict=True,
                            ),
                            interface_members=["start"],
                            children={},
                        ),
                        "sync": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.sync",
                            config=ModuleConfig(
                                path="tach.sync",
                                depends_on=[
                                    Dependency(path="tach.check"),
                                    Dependency(path="tach.errors"),
                                    Dependency(path="tach.filesystem"),
                                    Dependency(path="tach.parsing"),
                                ],
                                strict=True,
                            ),
                            interface_members=[
                                "sync_project",
                                "prune_dependency_constraints",
                            ],
                            children={},
                        ),
                        "test": ModuleNode(
                            is_end_of_path=True,
                            full_path="tach.test",
                            config=ModuleConfig(
                                path="tach.test",
                                depends_on=[
                                    Dependency(path="tach.errors"),
                                    Dependency(path="tach.filesystem"),
                                    Dependency(path="tach.filesystem.git_ops"),
                                    Dependency(path="tach.parsing"),
                                ],
                                strict=False,
                            ),
                            interface_members=[],
                            children={},
                        ),
                    },
                )
            },
        )
    )


@pytest.mark.parametrize(
    "changed_files,source_root,expected_mod_paths",
    [
        (["python/tach/test.py"], "python", ["tach.test"]),
        (["tach/test.py", "tach/a/test.py"], ".", ["tach.test", "tach.a.test"]),
        (["tach/a/__init__.py"], ".", ["tach.a"]),
    ],
)
def test_get_changed_module_paths(
    tmp_path, changed_files, source_root, expected_mod_paths
):
    project_root = tmp_path
    project_config = ProjectConfig(source_roots=[Path(source_root)])
    changed_files = [tmp_path / filepath for filepath in changed_files]
    assert set(expected_mod_paths) == set(
        get_changed_module_paths(project_root, project_config, changed_files)
    )


@pytest.mark.parametrize(
    "changed_mod_paths,expected_affected_modules",
    [
        (["tach.test"], ["tach.test", "tach.cli", "tach.__main__", "tach.start"]),
        (
            ["tach"],
            [
                "tach",
                "tach.cli",
                "tach.start",
                "tach.__main__",
                "tach.logging",
                "tach.cache",
            ],
        ),
        ([], []),
    ],
)
@patch("tach.test.get_changed_module_paths")
def test_affected_modules(
    mock_changed_module_paths,
    tmp_path,
    module_tree,
    modules,
    changed_mod_paths,
    expected_affected_modules,
):
    mock_changed_module_paths.return_value = changed_mod_paths
    project_root = tmp_path
    project_config = ProjectConfig(modules=modules)

    assert get_affected_modules(project_root, project_config, [], module_tree) == set(
        expected_affected_modules
    )



def get_project_imports(
    project_root: str,
    source_roots: list[str],
    file_path: str,
    ignore_type_checking_imports: bool,
) -> list[tuple[str, int]]: ...
def set_excluded_paths(exclude_paths: list[str]) -> None: ...
def create_dependency_report(
    project_root: str,
    source_roots: list[str],
    path: str,
    include_dependency_modules: list[str] | None,
    include_usage_modules: list[str] | None,
    skip_dependencies: bool,
    skip_usages: bool,
    ignore_type_checking_imports: bool,
) -> str: ...
def create_computation_cache_key(
    project_root: str,
    source_roots: list[str],
    action: str,
    py_interpreter_version: str,
    file_dependencies: list[str],
    env_dependencies: list[str],
    backend: str,
) -> str: ...
def check_computation_cache(
    project_root: str, cache_key: str
) -> tuple[list[tuple[int, str]], int] | None: ...
def update_computation_cache(
    project_root: str, cache_key: str, value: tuple[list[tuple[int, str]], int]
) -> None: ...



from __future__ import annotations

import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING

from tach import errors
from tach import filesystem as fs
from tach.extension import get_project_imports, set_excluded_paths
from tach.parsing import build_module_tree

if TYPE_CHECKING:
    from tach.core import (
        Dependency,  # noqa: TCH004
        ModuleNode,
        ModuleTree,
        ProjectConfig,
    )


@dataclass
class ErrorInfo:
    source_module: str = ""
    invalid_module: str = ""
    allowed_dependencies: list[Dependency] = field(default_factory=list)
    deprecated_dependencies: list[Dependency] = field(default_factory=list)
    exception_message: str = ""

    @property
    def is_dependency_error(self) -> bool:
        return all((self.source_module, self.invalid_module))

    @property
    def is_deprecated(self) -> bool:
        return self.is_dependency_error and self.invalid_module in [
            dep.path for dep in self.deprecated_dependencies
        ]


def is_top_level_module_import(mod_path: str, module: ModuleNode) -> bool:
    return mod_path == module.full_path


def import_matches_interface_members(mod_path: str, module: ModuleNode) -> bool:
    mod_path_segments = mod_path.rsplit(".", 1)
    if len(mod_path_segments) == 1:
        return mod_path_segments[0] == module.full_path
    else:
        mod_pkg_path, mod_member_name = mod_path_segments
        return (
            mod_pkg_path == module.full_path
            and mod_member_name in module.interface_members
        )


def check_import(
    module_tree: ModuleTree,
    import_mod_path: str,
    file_mod_path: str,
    file_nearest_module: ModuleNode | None = None,
) -> ErrorInfo | None:
    import_nearest_module = module_tree.find_nearest(import_mod_path)
    if import_nearest_module is None:
        # This shouldn't happen since we intend to filter out any external imports,
        # but we should allow external imports if they have made it here.
        return None

    # Lookup file_mod_path if module not given
    if file_nearest_module is None:
        file_nearest_module = module_tree.find_nearest(file_mod_path)
    # If module not found, we should fail since the implication is that
    # an external module is importing directly from our project
    if file_nearest_module is None:
        return ErrorInfo(
            exception_message=f"Module containing '{file_mod_path}' not found in project.",
        )

    # Imports within the same module are always allowed
    if import_nearest_module == file_nearest_module:
        return None

    import_module_config = import_nearest_module.config
    if import_module_config and import_module_config.strict:
        if not is_top_level_module_import(
            import_mod_path, import_nearest_module
        ) and not import_matches_interface_members(
            import_mod_path, import_nearest_module
        ):
            # In strict mode, import must be of the module itself or one of the
            # interface members (defined in __all__)
            return ErrorInfo(
                exception_message=(
                    f"Module '{import_nearest_module.full_path}' is in strict mode. "
                    "Only imports from the public interface of this module are allowed. "
                    f"The import '{import_mod_path}' (in module '{file_nearest_module.full_path}') "
                    f"is not included in __all__."
                ),
            )

    if not file_nearest_module.config or not import_nearest_module.config:
        return ErrorInfo(
            exception_message="Could not find module configuration.",
        )

    file_nearest_module_path = file_nearest_module.config.path
    import_nearest_module_path = import_nearest_module.config.path

    # The import must be explicitly allowed
    dependencies = file_nearest_module.config.depends_on

    allowed_dependencies = [dep for dep in dependencies if not dep.deprecated]
    deprecated_dependencies = [dep for dep in dependencies if dep.deprecated]
    if any(dep.path == import_nearest_module_path for dep in allowed_dependencies):
        # The import matches at least one expected dependency
        return None
    if any(dep.path == import_nearest_module_path for dep in deprecated_dependencies):
        # Dependency exists but is deprecated
        return ErrorInfo(
            source_module=file_nearest_module_path,
            invalid_module=import_nearest_module_path,
            allowed_dependencies=allowed_dependencies,
            deprecated_dependencies=deprecated_dependencies,
        )
    # This means the import is not declared as a dependency of the file
    return ErrorInfo(
        source_module=file_nearest_module_path,
        invalid_module=import_nearest_module_path,
        allowed_dependencies=allowed_dependencies,
        deprecated_dependencies=deprecated_dependencies,
    )


@dataclass
class BoundaryError:
    file_path: Path
    line_number: int
    import_mod_path: str
    error_info: ErrorInfo


@dataclass
class CheckResult:
    errors: list[BoundaryError] = field(default_factory=list)
    deprecated_warnings: list[BoundaryError] = field(default_factory=list)
    warnings: list[str] = field(default_factory=list)


def is_path_excluded(path: Path, exclude_paths: list[str]) -> bool:
    dirpath_for_matching = f"{path}/"
    return any(
        re.match(exclude_path, dirpath_for_matching) for exclude_path in exclude_paths
    )


def check(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str] | None = None,
) -> CheckResult:
    if not project_root.is_dir():
        raise errors.TachSetupError(
            f"The path {project_root} is not a valid directory."
        )

    boundary_errors: list[BoundaryError] = []
    boundary_warnings: list[BoundaryError] = []
    warnings: list[str] = []

    if exclude_paths is not None and project_config.exclude is not None:
        exclude_paths.extend(project_config.exclude)
    else:
        exclude_paths = project_config.exclude

    source_roots = [
        project_root / source_root for source_root in project_config.source_roots
    ]

    module_validation_result = fs.validate_project_modules(
        source_roots=source_roots, modules=project_config.modules
    )
    warnings.extend(
        f"Module '{module.path}' not found. It will be ignored."
        for module in module_validation_result.invalid_modules
    )
    module_tree = build_module_tree(
        source_roots=source_roots,
        modules=module_validation_result.valid_modules,
        forbid_circular_dependencies=project_config.forbid_circular_dependencies,
    )

    found_at_least_one_project_import = False
    # This informs the Rust extension ahead-of-time which paths are excluded.
    # The extension builds regexes and uses them during `get_project_imports`
    set_excluded_paths(exclude_paths=exclude_paths or [])
    for source_root in source_roots:
        for file_path in fs.walk_pyfiles(source_root):
            abs_file_path = source_root / file_path
            rel_file_path = abs_file_path.relative_to(project_root)
            if is_path_excluded(rel_file_path, exclude_paths=exclude_paths or []):
                continue

            mod_path = fs.file_to_module_path(
                source_roots=tuple(source_roots), file_path=abs_file_path
            )
            nearest_module = module_tree.find_nearest(mod_path)
            if nearest_module is None:
                continue

            try:
                project_imports = get_project_imports(
                    project_root=str(project_root),
                    source_roots=list(map(str, source_roots)),
                    file_path=str(abs_file_path),
                    ignore_type_checking_imports=project_config.ignore_type_checking_imports,
                )
            except SyntaxError:
                warnings.append(f"Skipping '{file_path}' due to a syntax error.")
                continue
            except OSError:
                warnings.append(f"Skipping '{file_path}' due to a file system error.")
                continue
            for project_import in project_imports:
                found_at_least_one_project_import = True
                error_info = check_import(
                    module_tree=module_tree,
                    import_mod_path=project_import[0],
                    file_nearest_module=nearest_module,
                    file_mod_path=mod_path,
                )
                if error_info is None:
                    continue
                boundary_error = BoundaryError(
                    file_path=file_path,
                    import_mod_path=project_import[0],
                    line_number=project_import[1],
                    error_info=error_info,
                )
                if error_info.is_deprecated:
                    boundary_warnings.append(boundary_error)
                else:
                    boundary_errors.append(boundary_error)

    if not found_at_least_one_project_import:
        warnings.append(
            "WARNING: No first-party imports were found. You may need to use 'tach mod' to update your Python source roots. Docs: https://docs.gauge.sh/usage/configuration#source-roots"
        )
    return CheckResult(
        errors=boundary_errors, deprecated_warnings=boundary_warnings, warnings=warnings
    )


__all__ = ["BoundaryError", "check"]



from __future__ import annotations

from pathlib import Path

import pytest

from tach.core import (
    Dependency,
    ModuleConfig,
    ModuleNode,
    ModuleTree,
)


@pytest.fixture
def example_dir() -> Path:
    current_dir = Path(__file__).parent
    return current_dir / "example"


@pytest.fixture
def test_config() -> ModuleConfig:
    return ModuleConfig(path="test", strict=False)


@pytest.fixture
def module_tree() -> ModuleTree:
    return ModuleTree(
        root=ModuleNode(
            is_end_of_path=False,
            full_path="",
            config=None,
            children={
                "domain_one": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_one",
                    config=ModuleConfig(
                        path="domain_one",
                        depends_on=[
                            Dependency(path="domain_one.subdomain", deprecated=True),
                            Dependency(path="domain_three"),
                        ],
                        strict=True,
                    ),
                    interface_members=["public_fn"],
                    children={
                        "subdomain": ModuleNode(
                            is_end_of_path=True,
                            full_path="domain_one.subdomain",
                            config=ModuleConfig(
                                path="domain_one.subdomain", strict=True
                            ),
                            children={},
                        )
                    },
                ),
                "domain_two": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_two",
                    config=ModuleConfig(
                        path="domain_two",
                        depends_on=[Dependency(path="domain_one")],
                        strict=False,
                    ),
                    children={
                        "subdomain": ModuleNode(
                            is_end_of_path=True,
                            full_path="domain_two.subdomain",
                            config=ModuleConfig(
                                path="domain_two",
                                depends_on=[Dependency(path="domain_one")],
                                strict=False,
                            ),
                            children={},
                        )
                    },
                ),
                "domain_three": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_three",
                    config=ModuleConfig(path="domain_three", strict=False),
                    children={},
                ),
            },
        )
    )



from __future__ import annotations

import os
from pathlib import Path

import pytest
from git import Repo

from tach.filesystem.git_ops import get_changed_files


@pytest.fixture
def git_repo(tmp_path):
    os.environ["GIT_AUTHOR_NAME"] = "Temporary User"
    os.environ["GIT_AUTHOR_EMAIL"] = "temporary.user@example.com"
    os.environ["GIT_COMMITTER_NAME"] = "Temporary User"
    os.environ["GIT_COMMITTER_EMAIL"] = "temporary.user@example.com"

    repo_path = tmp_path / "test_repo"
    repo = Repo.init(repo_path, initial_branch="main")

    files = {
        "file1.txt": "Hello, World!",
        "dir1/file2.txt": "This is file 2.",
        "dir2/file3.txt": "This is file 3.",
    }

    for file_path, content in files.items():
        full_path = repo_path / file_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.write_text(content)

    repo.git.add("-A")
    repo.git.commit("-m", "Initial commit on main branch")

    yield repo_path

    del os.environ["GIT_AUTHOR_NAME"]
    del os.environ["GIT_AUTHOR_EMAIL"]
    del os.environ["GIT_COMMITTER_NAME"]
    del os.environ["GIT_COMMITTER_EMAIL"]


@pytest.mark.parametrize(
    "setup_changes, expected_files",
    [
        (lambda repo_path: None, set()),
        (
            lambda repo_path: {
                (repo_path / "dir1/file4.txt").write_text("This is file 4."),
                (repo_path / "dir2/file5.txt").write_text("This is file 5."),
            },
            {"dir1/file4.txt", "dir2/file5.txt"},
        ),
        (
            lambda repo_path: {
                (repo_path / "file1.txt").unlink(),
                (repo_path / "dir1/file2.txt").unlink(),
            },
            {"file1.txt", "dir1/file2.txt"},
        ),
        (
            lambda repo_path: {
                (repo_path / "file1.txt").rename(repo_path / "file1_renamed.txt"),
                (repo_path / "dir1/file2.txt").rename(
                    repo_path / "dir1/file2_renamed.txt"
                ),
            },
            {
                "file1.txt",
                "file1_renamed.txt",
                "dir1/file2.txt",
                "dir1/file2_renamed.txt",
            },
        ),
    ],
)
def test_changed_files_new_branch(git_repo, setup_changes, expected_files):
    repo_path = git_repo
    repo = Repo(repo_path)

    repo.git.checkout("-b", "new_branch")

    if setup_changes:
        setup_changes(repo_path)

    repo.git.add("-A")
    repo.git.commit("-m", "Changes on new_branch", "--allow-empty")

    changed_files = get_changed_files(repo_path, base="main", head="new_branch")

    assert set(
        changed_file.relative_to(git_repo) for changed_file in changed_files
    ) == set(Path(filepath) for filepath in expected_files)


@pytest.mark.parametrize(
    "setup_changes, expected_files",
    [
        (lambda repo_path: None, set()),
        (
            lambda repo_path: {
                (repo_path / "dir1/file4.txt").write_text("This is file 4."),
                (repo_path / "dir2/file5.txt").write_text("This is file 5."),
            },
            {"dir1/file4.txt", "dir2/file5.txt"},
        ),
        (
            lambda repo_path: {
                (repo_path / "file1.txt").unlink(),
                (repo_path / "dir1/file2.txt").unlink(),
            },
            {"file1.txt", "dir1/file2.txt"},
        ),
        (
            lambda repo_path: {
                (repo_path / "file1.txt").rename(repo_path / "file1_renamed.txt"),
                (repo_path / "dir1/file2.txt").rename(
                    repo_path / "dir1/file2_renamed.txt"
                ),
            },
            {
                "file1.txt",
                "file1_renamed.txt",
                "dir1/file2.txt",
                "dir1/file2_renamed.txt",
            },
        ),
    ],
)
def test_changed_files_working_directory(git_repo, setup_changes, expected_files):
    if setup_changes:
        setup_changes(git_repo)

    changed_files = get_changed_files(git_repo, base="main")

    assert set(
        changed_file.relative_to(git_repo) for changed_file in changed_files
    ) == set(Path(filepath) for filepath in expected_files)



from __future__ import annotations

from tach.core.config import (
    Dependency,
    ModuleConfig,
    ProjectConfig,
    UnusedDependencies,
)
from tach.core.modules import ModuleNode, ModuleTree

__all__ = [
    "ProjectConfig",
    "ModuleConfig",
    "ModuleNode",
    "ModuleTree",
    "Dependency",
    "UnusedDependencies",
]



from __future__ import annotations

from pathlib import Path
from unittest.mock import Mock

import pytest

from tach import cli
from tach.check import BoundaryError, CheckResult, ErrorInfo
from tach.core import Dependency, ModuleConfig, ProjectConfig


@pytest.fixture
def mock_check(mocker) -> Mock:
    mock = Mock(return_value=CheckResult())  # default to a return with no errors
    mocker.patch("tach.cli.check", mock)
    return mock


@pytest.fixture
def mock_project_config(mocker) -> None:
    def mock_project_config(root: str = "") -> ProjectConfig:
        return ProjectConfig(
            modules=[
                ModuleConfig(path="mocked", depends_on=[Dependency(path="mocked")])
            ]
        )

    mocker.patch("tach.cli.parse_project_config", mock_project_config)


def test_execute_with_tach_yml(capfd, mock_check, mock_project_config):
    # Test with a valid path as mocked
    args, _ = cli.parse_arguments(["check"])
    assert args.command == "check"
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check(Path())
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All module dependencies validated!" in captured.out


def test_execute_with_error(capfd, mock_check, mock_project_config):
    # Mock an error returned from check
    location = Path("valid_dir/file.py")
    message = "Import valid_dir in valid_dir/file.py is blocked by boundary"
    mock_check.return_value = CheckResult(
        errors=[
            BoundaryError(
                file_path=location,
                line_number=0,
                import_mod_path="valid_dir",
                error_info=ErrorInfo(
                    exception_message="Import valid_dir in valid_dir/file.py is blocked by boundary",
                ),
            )
        ]
    )
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check(Path())
    captured = capfd.readouterr()
    assert sys_exit.value.code == 1
    assert str(location) in captured.err
    assert message in captured.err


def test_invalid_command(capfd):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with an invalid command
        cli.parse_arguments(["help"])
    captured = capfd.readouterr()
    assert sys_exit.value.code == 2
    assert "invalid choice: 'help" in captured.err


def test_execute_with_valid_exclude(capfd, mock_check, mock_project_config):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with a valid path as mocked
        args, _ = cli.parse_arguments(["check", "--exclude", "valid_dir"])
        exclude_paths = args.exclude.split(",")
        cli.tach_check(Path(), exclude_paths=exclude_paths)
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All module dependencies validated!" in captured.out



from __future__ import annotations

from typing import TYPE_CHECKING

import networkx as nx

from tach.core import ModuleTree
from tach.errors import TachCircularDependencyError
from tach.parsing import parse_interface_members

if TYPE_CHECKING:
    from pathlib import Path

    from tach.core import ModuleConfig


def find_duplicate_modules(modules: list[ModuleConfig]) -> list[str]:
    duplicate_module_paths: list[str] = []
    seen: set[str] = set()
    for module in modules:
        if module.path in seen:
            duplicate_module_paths.append(module.path)
            continue
        seen.add(module.path)
    return duplicate_module_paths


def canonical_form(cycle: list[str]) -> list[str]:
    """Rotate cycle to start with the smallest element."""
    min_index = cycle.index(min(cycle))
    return cycle[min_index:] + cycle[:min_index]


def find_cycles(
    modules: list[ModuleConfig],
) -> list[list[str]]:
    graph = nx.DiGraph()  # type: ignore
    # Add nodes
    for module in modules:
        graph.add_node(module.path)  # type: ignore

    # Add dependency edges
    for module in modules:
        for dependency in module.depends_on:
            graph.add_edge(module.path, dependency.path)  # type: ignore

    all_cycles: list[list[str]] = list(nx.simple_cycles(graph))  # type: ignore

    canonical_cycles = {tuple(canonical_form(cycle)) for cycle in all_cycles}

    unique_cycles = [list(cycle) for cycle in canonical_cycles]

    return unique_cycles


def build_module_tree(
    source_roots: list[Path],
    modules: list[ModuleConfig],
    forbid_circular_dependencies: bool,
) -> ModuleTree:
    duplicate_modules = find_duplicate_modules(modules)
    if duplicate_modules:
        raise ValueError(
            f"Failed to build module tree. The following modules were defined more than once: {duplicate_modules}"
        )
    if forbid_circular_dependencies:
        cycles = find_cycles(modules)
        if cycles:
            raise TachCircularDependencyError(cycles)
    tree = ModuleTree()
    for module in modules:
        tree.insert(
            config=module,
            path=module.mod_path,
            interface_members=parse_interface_members(
                source_roots=source_roots, module_path=module.path
            ),
        )
    return tree



from __future__ import annotations

from pathlib import Path

from git import GitCommandError, InvalidGitRepositoryError, NoSuchPathError, Repo

from tach.errors import TachError, TachSetupError


def get_changed_files(
    project_root: Path, head: str = "", base: str = "main"
) -> list[Path]:
    try:
        repo = Repo(project_root, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        raise TachSetupError(
            "The project does not appear to be a git repository, cannot determine changed files!"
        )

    try:
        if head:
            diff: str = repo.git.diff("--name-status", head, base)
        else:
            # If head is not provided, we can diff against 'base' from the current filesystem
            diff: str = repo.git.diff("--name-status", base)
    except GitCommandError:
        head_display = f"'{head}'" if head else "current filesystem"
        raise TachError(f"Failed to check diff between '{base}' and {head_display}!")

    diff_lines = diff.splitlines()
    changed_files: set[str] = set()

    for line in diff_lines:
        _, *files = line.split("\t")
        changed_files.update(files)

    if not head:
        # If we are using the current filesystem, there may be relevant changes in untracked files
        untracked_files: str = repo.git.ls_files("--others", "--exclude-standard")
        changed_files.update(untracked_files.splitlines())

    # return list of unique Paths
    git_root: str = repo.git.rev_parse("--show-toplevel")
    return [(Path(git_root) / filepath).resolve() for filepath in changed_files]


__all__ = ["get_changed_files"]



from __future__ import annotations

from tach.parsing.config import (
    dump_project_config_to_yaml,
    parse_project_config,
)
from tach.parsing.interface import parse_interface_members
from tach.parsing.modules import build_module_tree, find_cycles

__all__ = [
    "parse_project_config",
    "dump_project_config_to_yaml",
    "parse_interface_members",
    "build_module_tree",
    "find_cycles",
]



from __future__ import annotations

from copy import copy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from tach import filesystem as fs
from tach.errors import TachError, TachSetupError
from tach.extension import get_project_imports
from tach.filesystem.git_ops import get_changed_files
from tach.parsing import build_module_tree

if TYPE_CHECKING:
    from pathlib import Path

    from tach.core import ModuleConfig, ModuleTree, ProjectConfig


def build_module_consumer_map(modules: list[ModuleConfig]) -> dict[str, list[str]]:
    consumer_map: dict[str, list[str]] = {}
    for module in modules:
        for dependency in module.depends_on:
            if dependency.path in consumer_map:
                consumer_map[dependency.path].append(module.mod_path)
            else:
                consumer_map[dependency.path] = [module.mod_path]
    return consumer_map


def get_changed_module_paths(
    project_root: Path, project_config: ProjectConfig, changed_files: list[Path]
) -> list[str]:
    source_roots = [
        project_root / source_root for source_root in project_config.source_roots
    ]
    changed_module_paths = [
        fs.file_to_module_path(source_roots=tuple(source_roots), file_path=changed_file)
        for changed_file in changed_files
        if any(source_root in changed_file.parents for source_root in source_roots)
        and changed_file.suffix == ".py"
    ]

    return changed_module_paths


def find_affected_modules(
    root_module_path: str,
    module_consumers: dict[str, list[str]],
    known_affected_modules: set[str],
) -> set[str]:
    if root_module_path not in module_consumers:
        return known_affected_modules
    for consumer in module_consumers[root_module_path]:
        # avoid recursing on modules we have already seen to prevent infinite cycles
        if consumer not in known_affected_modules:
            known_affected_modules.add(consumer)
            known_affected_modules |= find_affected_modules(
                consumer,
                module_consumers=module_consumers,
                known_affected_modules=known_affected_modules,
            )
    return known_affected_modules


def get_affected_modules(
    project_root: Path,
    project_config: ProjectConfig,
    changed_files: list[Path],
    module_tree: ModuleTree,
) -> set[str]:
    changed_module_paths = get_changed_module_paths(
        project_root, project_config, changed_files
    )
    affected_modules: set[str] = set()
    for changed_mod_path in changed_module_paths:
        nearest_module = module_tree.find_nearest(changed_mod_path)
        if nearest_module is None:
            raise TachError(
                f"Could not find module containing path: {changed_mod_path}"
            )
        affected_modules.add(nearest_module.full_path)

    module_consumers = build_module_consumer_map(project_config.modules)
    for module in list(affected_modules):
        find_affected_modules(
            module,
            module_consumers=module_consumers,
            known_affected_modules=affected_modules,
        )
    return affected_modules


@dataclass
class AffectedTestsResult:
    exit_code: int
    tests_ran_to_completion: bool


def run_affected_tests(
    project_root: Path,
    project_config: ProjectConfig,
    head: str,
    base: str,
    pytest_args: list[Any] | None = None,
) -> AffectedTestsResult:
    try:
        import pytest  # type: ignore  # noqa: F401
    except ImportError:
        raise TachSetupError("Cannot run tests, could not find 'pytest'.")

    class TachPytestPlugin:
        def __init__(
            self,
            project_root: Path,
            source_roots: list[Path],
            module_tree: ModuleTree,
            affected_modules: set[str],
            all_affected_files: set[Path],
        ):
            self.project_root = project_root
            self.source_roots = source_roots
            self.module_tree = module_tree
            self.affected_modules = affected_modules
            self.all_affected_files = all_affected_files
            self.removed_test_paths: set[Path] = set()
            self.num_removed_items: int = 0
            self.tests_ran_to_completion = False

        def pytest_collection_modifyitems(
            self,
            session: pytest.Session,
            config: pytest.Config,
            items: list[pytest.Item],
        ):
            seen: set[Path] = set()
            for item in copy(items):
                if not item.path:
                    continue
                if item.path in self.removed_test_paths:
                    self.num_removed_items += 1
                    items.remove(item)
                    continue
                if item.path in seen:
                    continue

                if item.path in self.all_affected_files:
                    # If this test file was changed,
                    # then we know we need to rerun it
                    seen.add(item.path)
                    continue

                project_imports = get_project_imports(
                    project_root=str(self.project_root),
                    source_roots=list(map(str, self.source_roots)),
                    file_path=str(item.path.resolve()),
                    ignore_type_checking_imports=True,
                )
                for mod_path, _ in project_imports:
                    nearest_module = self.module_tree.find_nearest(mod_path)
                    if not nearest_module:
                        continue
                    if nearest_module.full_path in self.affected_modules:
                        # We can break early without any modifications, since we know this file path is affected
                        break
                else:
                    # If none of the project imports in the test are affected, we can skip all tests in the file path
                    self.num_removed_items += 1
                    items.remove(item)
                    self.removed_test_paths.add(item.path)
                seen.add(item.path)

        def pytest_report_collectionfinish(
            self,
            config: pytest.Config,
            start_path: Path,
            startdir: Any,
            items: list[pytest.Item],
        ) -> str | list[str]:
            return [
                f"[Tach] Skipped {len(self.removed_test_paths)} test file{'s' if len(self.removed_test_paths) > 1 else ''}"
                f" ({self.num_removed_items} tests)"
                " since they were unaffected by current changes.",
                *(
                    f"[Tach] > Skipped '{test_path}'"
                    for test_path in self.removed_test_paths
                ),
            ]

        def pytest_terminal_summary(
            self, terminalreporter: Any, exitstatus: int, config: pytest.Config
        ):
            self.tests_ran_to_completion = True

    source_roots = [
        project_root / source_root for source_root in project_config.source_roots
    ]

    module_validation_result = fs.validate_project_modules(
        source_roots=source_roots, modules=project_config.modules
    )
    # TODO: log warning
    for module in module_validation_result.invalid_modules:
        print(f"Module '{module.path}' not found. It will be ignored.")

    module_tree = build_module_tree(
        source_roots=source_roots,
        modules=module_validation_result.valid_modules,
        forbid_circular_dependencies=project_config.forbid_circular_dependencies,
    )

    changed_files = get_changed_files(project_root, head=head, base=base)
    affected_module_paths = get_affected_modules(
        project_root,
        project_config,
        changed_files=changed_files,
        module_tree=module_tree,
    )
    pytest_plugin = TachPytestPlugin(
        project_root=project_root,
        source_roots=source_roots,
        module_tree=module_tree,
        affected_modules=affected_module_paths,
        all_affected_files={changed_file.resolve() for changed_file in changed_files},
    )

    exit_code = pytest.main(pytest_args, plugins=[pytest_plugin])

    if exit_code == pytest.ExitCode.NO_TESTS_COLLECTED:
        # Selective testing means running zero tests will happen regularly,
        # so we do not want the default behavior of failing when no tests
        # are collected.
        exit_code = pytest.ExitCode.OK

    return AffectedTestsResult(
        exit_code=exit_code,
        tests_ran_to_completion=pytest_plugin.tests_ran_to_completion,
    )


__all__ = ["run_affected_tests"]



from __future__ import annotations

import os
from pathlib import Path

import pytest

from tach.core.config import ProjectConfig
from tach.errors import TachError
from tach.report import report


@pytest.fixture
def mock_project_config() -> ProjectConfig:
    return ProjectConfig()


@pytest.fixture
def mock_cwd(tmp_path):
    try:
        original_path = os.getcwd()
        os.chdir(tmp_path)
        yield tmp_path
    finally:
        os.chdir(original_path)


# The code assumes that the cwd is within the project root
# due to pre-condition checks


def test_valid_path(mock_project_config, mock_cwd):
    mock_path = mock_cwd / "test.py"
    mock_path.touch()
    result = report(
        project_root=mock_cwd, path=Path("test.py"), project_config=mock_project_config
    )
    assert result


def test_valid_dir(mock_project_config, mock_cwd):
    mock_path = mock_cwd / "test"
    mock_path.mkdir()
    result = report(
        project_root=mock_cwd, path=Path("test"), project_config=mock_project_config
    )
    assert result


def test_valid_dir_trailing_slash(mock_project_config, mock_cwd):
    mock_path = mock_cwd / "test"
    mock_path.mkdir()
    result = report(
        project_root=mock_cwd,
        path=Path("test/"),
        project_config=mock_project_config,
    )
    assert result


def test_invalid_root(mock_project_config, mock_cwd):
    with pytest.raises(TachError):
        report(
            project_root=Path("Invalid!!"),
            path=Path("."),
            project_config=mock_project_config,
        )


def test_invalid_path(mock_project_config, mock_cwd):
    with pytest.raises(TachError):
        report(
            project_root=mock_cwd,
            path=Path("Invalid!!"),
            project_config=mock_project_config,
        )



from __future__ import annotations

from pathlib import Path, PosixPath

import pytest
from pydantic import ValidationError

from tach.constants import DEFAULT_EXCLUDE_PATHS, ROOT_MODULE_SENTINEL_TAG
from tach.core import Dependency, ModuleConfig, ProjectConfig
from tach.core.config import CacheConfig
from tach.filesystem import file_to_module_path
from tach.parsing import find_cycles, parse_project_config


@pytest.fixture
def example_dir() -> Path:
    current_dir = Path(__file__).parent
    return current_dir / "example"


def test_file_to_mod_path():
    assert file_to_module_path((Path("."),), Path("__init__.py")) == "."
    assert (
        file_to_module_path((Path("."),), Path("domain_one", "__init__.py"))
        == "domain_one"
    )
    assert (
        file_to_module_path((Path("."),), Path("domain_one", "interface.py"))
        == "domain_one.interface"
    )
    assert (
        file_to_module_path((Path("source", "root"),), Path("source", "root", "domain"))
        == "domain"
    )
    assert (
        file_to_module_path(
            (Path("src1"), Path("src2")), Path("src1", "core", "lib", "cat.py")
        )
        == "core.lib.cat"
    )


def test_parse_valid_project_config(example_dir):
    result = parse_project_config(example_dir / "valid")
    assert result == ProjectConfig(
        modules=[
            ModuleConfig(
                path="domain_one",
                depends_on=[Dependency(path="domain_two", deprecated=True)],
                strict=False,
            ),
            ModuleConfig(path="domain_three", depends_on=[], strict=False),
            ModuleConfig(
                path="domain_two",
                depends_on=[Dependency(path="domain_three", deprecated=False)],
                strict=False,
            ),
            ModuleConfig(
                path=ROOT_MODULE_SENTINEL_TAG,
                depends_on=[Dependency(path="domain_one", deprecated=False)],
                strict=False,
            ),
        ],
        cache=CacheConfig(backend="local", file_dependencies=[], env_dependencies=[]),
        exclude=sorted(DEFAULT_EXCLUDE_PATHS),
        source_roots=[PosixPath(".")],
        exact=True,
        disable_logging=False,
        ignore_type_checking_imports=True,
        forbid_circular_dependencies=True,
    )


def test_invalid_project_config(example_dir):
    with pytest.raises(ValidationError):
        parse_project_config(example_dir / "invalid")


def test_empty_project_config(example_dir):
    with pytest.raises(ValueError):
        parse_project_config(example_dir / "invalid" / "empty")


def test_valid_circular_dependencies(example_dir):
    project_config = parse_project_config(example_dir / "valid")
    modules = project_config.modules
    all_cycles: list[list[str]] = find_cycles(modules)
    assert all_cycles == []


def test_cycles_circular_dependencies(example_dir):
    project_config = parse_project_config(example_dir / "cycles")
    modules = project_config.modules
    all_cycles: list[list[str]] = find_cycles(modules)
    tuple_cycles: list[tuple[str]] = [tuple(cycle) for cycle in all_cycles]
    assert set(tuple_cycles) == {
        ("domain_one", "domain_two", "domain_three"),
        ("domain_one", "domain_three"),
    }



from __future__ import annotations

import pytest

from tach.core import ModuleConfig, ModuleNode, ModuleTree


@pytest.fixture
def test_config() -> ModuleConfig:
    return ModuleConfig(path="test", strict=False)


@pytest.fixture
def module_tree() -> ModuleTree:
    return ModuleTree(
        root=ModuleNode(
            is_end_of_path=True,
            full_path=".",
            config=ModuleConfig(path="<root>"),
            children={
                "domain_one": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_one",
                    config=ModuleConfig(path="test", strict=False),
                    children={
                        "subdomain": ModuleNode(
                            is_end_of_path=True,
                            full_path="domain_one.subdomain",
                            config=ModuleConfig(path="test", strict=False),
                            children={},
                        )
                    },
                ),
                "domain_two": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_two",
                    config=ModuleConfig(path="test", strict=False),
                    children={
                        "subdomain": ModuleNode(
                            is_end_of_path=True,
                            full_path="domain_two.subdomain",
                            config=ModuleConfig(path="test", strict=False),
                            children={},
                        )
                    },
                ),
                "domain_three": ModuleNode(
                    is_end_of_path=True,
                    full_path="domain_three",
                    config=ModuleConfig(path="test", strict=False),
                    children={},
                ),
            },
        )
    )


def test_iterate_over_empty_tree():
    tree = ModuleTree()
    assert list(tree) == [tree.root]


def test_iterate_over_populated_tree(module_tree):
    assert set(node.full_path for node in module_tree) == {
        ".",
        "domain_one",
        "domain_one.subdomain",
        "domain_two",
        "domain_two.subdomain",
        "domain_three",
    }


def test_get_nonexistent_path(module_tree):
    assert module_tree.get("fakepath") is None


def test_get_empty_path():
    tree = ModuleTree()
    assert tree.get("") is None


def test_get_actual_path(module_tree):
    assert module_tree.get("domain_one") is not None


def test_insert_empty_path(test_config):
    tree = ModuleTree()
    with pytest.raises(ValueError):
        tree.insert(test_config, "", [])


def test_insert_single_level_path(test_config):
    tree = ModuleTree()
    tree.insert(test_config, "domain", [])
    assert set(node.full_path for node in tree) == {".", "domain"}


def test_insert_multi_level_path(test_config):
    tree = ModuleTree()
    tree.insert(test_config, "domain.subdomain", [])
    assert set(node.full_path for node in tree) == {".", "domain.subdomain"}


def test_find_nearest_at_root(module_tree):
    # ModuleTree is not responsible for filtering to project imports
    module = module_tree.find_nearest("other_domain")
    assert module is module_tree.root


def test_find_nearest_in_single_domain(module_tree):
    module = module_tree.find_nearest("domain_one.thing")
    assert module.full_path == "domain_one"


def test_find_nearest_in_nested_domain(module_tree):
    module = module_tree.find_nearest("domain_two.subdomain.thing")
    assert module.full_path == "domain_two.subdomain"



from __future__ import annotations

from typing import TYPE_CHECKING

from tach import errors
from tach import filesystem as fs
from tach.check import check
from tach.core import Dependency
from tach.filesystem import get_project_config_path
from tach.parsing import dump_project_config_to_yaml

if TYPE_CHECKING:
    from pathlib import Path

    from tach.core import ProjectConfig


def sync_dependency_constraints(
    project_root: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str] | None = None,
    prune: bool = True,
) -> ProjectConfig:
    """
    Update project configuration with auto-detected dependency constraints.
    If prune is set to False, it will create dependencies to resolve existing errors,
    but will not remove any constraints.
    """
    deprecation_map: dict[str, list[str]] = {}
    if prune:
        # Create a blank config
        new_config = project_config.model_copy(
            update={
                "modules": [
                    module.model_copy(update={"depends_on": []})
                    for module in project_config.modules
                ]
            }
        )
        # Find deprecations - only needed if pruning as otherwise they will not be removed
        for module in project_config.modules:
            for dependency in module.depends_on:
                if dependency.deprecated:
                    if module.path not in deprecation_map:
                        deprecation_map[module.path] = [dependency.path]
                    else:
                        deprecation_map[module.path].append(dependency.path)
    else:
        # Use the same config, existing deprecations will remain
        new_config = project_config
    check_result = check(
        project_root=project_root,
        project_config=new_config,
        exclude_paths=exclude_paths,
    )
    for error in check_result.errors:
        error_info = error.error_info
        if error_info.is_dependency_error:
            source_path = error_info.source_module
            dep_path = error_info.invalid_module
            deprecated = (
                source_path in deprecation_map
                and dep_path in deprecation_map[error_info.source_module]
            )
            if deprecated:
                dependency = Dependency(
                    path=error_info.invalid_module, deprecated=deprecated
                )
            else:
                dependency = Dependency(path=error_info.invalid_module)

            new_config.add_dependency_to_module(
                module=source_path, dependency=dependency
            )
    return new_config


def sync_project(
    project_root: Path,
    project_config: ProjectConfig,
    add: bool = False,
    exclude_paths: list[str] | None = None,
) -> None:
    tach_yml_path = get_project_config_path(project_root)
    if tach_yml_path is None:
        raise errors.TachError(
            "Unexpected error. Could not find configuration file during 'sync'."
        )

    project_config = sync_dependency_constraints(
        project_root=project_root,
        project_config=project_config,
        exclude_paths=exclude_paths,
        prune=not add,
    )
    tach_yml_content = dump_project_config_to_yaml(project_config)
    fs.write_file(str(tach_yml_path), tach_yml_content)


__all__ = ["sync_project", "sync_dependency_constraints"]



from __future__ import annotations

import argparse
import os
import sys
from dataclasses import dataclass, field
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any

from tach import __version__, cache
from tach import filesystem as fs
from tach.check import BoundaryError, check
from tach.colors import BCOLORS
from tach.constants import CONFIG_FILE_NAME, TOOL_NAME
from tach.core import ProjectConfig
from tach.errors import TachCircularDependencyError, TachError
from tach.extension import (
    check_computation_cache,
    create_computation_cache_key,
    update_computation_cache,
)
from tach.filesystem import install_pre_commit
from tach.logging import LogDataModel, logger
from tach.mod import mod_edit_interactive
from tach.parsing import parse_project_config
from tach.report import report
from tach.show import generate_module_graph_dot_file, generate_show_url
from tach.sync import (
    sync_dependency_constraints,
    sync_project,
)
from tach.test import run_affected_tests

if TYPE_CHECKING:
    from tach.core import UnusedDependencies


class TerminalEnvironment(Enum):
    UNKNOWN = 1
    JETBRAINS = 2
    VSCODE = 3


@lru_cache(maxsize=None)
def detect_environment() -> TerminalEnvironment:
    if "jetbrains" in os.environ.get("TERMINAL_EMULATOR", "").lower():
        return TerminalEnvironment.JETBRAINS
    elif "vscode" in os.environ.get("TERM_PROGRAM", "").lower():
        return TerminalEnvironment.VSCODE
    return TerminalEnvironment.UNKNOWN


def create_clickable_link(
    file_path: Path, display_path: Path | None = None, line: int | None = None
) -> str:
    terminal_env = detect_environment()
    abs_path = file_path.resolve()

    if terminal_env == TerminalEnvironment.JETBRAINS:
        link = f"file://{abs_path}:{line}" if line is not None else f"file://{abs_path}"
    elif terminal_env == TerminalEnvironment.VSCODE:
        link = (
            f"vscode://file/{abs_path}:{line}"
            if line is not None
            else f"vscode://file/{abs_path}"
        )
    else:
        # For generic terminals, use a standard file link
        link = f"file://{abs_path}"

    # ANSI escape codes for clickable link
    if line:
        # Show the line number if we have it
        display_file_path = f"{display_path or file_path}[L{line}]"
    else:
        display_file_path = str(display_path) if display_path else str(file_path)
    clickable_link = f"\033]8;;{link}\033\\{display_file_path}\033]8;;\033\\"
    return clickable_link


def build_error_message(error: BoundaryError, source_roots: list[Path]) -> str:
    absolute_error_path = next(
        (
            source_root / error.file_path
            for source_root in source_roots
            if (source_root / error.file_path).exists()
        ),
        None,
    )

    if absolute_error_path is None:
        # This is an unexpected case,
        # all errors should have originated from within a source root
        error_location = error.file_path
    else:
        error_location = create_clickable_link(
            absolute_error_path,
            display_path=error.file_path,
            line=error.line_number,
        )

    error_template = (
        f"❌ {BCOLORS.FAIL}{error_location}{BCOLORS.ENDC}{BCOLORS.FAIL}: "
        f"{{message}} {BCOLORS.ENDC}"
    )
    warning_template = (
        f"‼️ {BCOLORS.FAIL}{error_location}{BCOLORS.ENDC}{BCOLORS.WARNING}: "
        f"{{message}} {BCOLORS.ENDC}"
    )
    error_info = error.error_info
    if error_info.exception_message:
        return error_template.format(message=error_info.exception_message)
    elif not error_info.is_dependency_error:
        return error_template.format(message="Unexpected error")

    error_message = (
        f"Cannot import '{error.import_mod_path}'. "
        f"'{error_info.source_module}' cannot depend on '{error_info.invalid_module}'."
    )

    warning_message = (
        f"Import '{error.import_mod_path}' is deprecated. "
        f"'{error_info.source_module}' should not depend on '{error_info.invalid_module}'."
    )
    if error_info.is_deprecated:
        return warning_template.format(message=warning_message)
    return error_template.format(message=error_message)


def print_warnings(warning_list: list[str]) -> None:
    for warning in warning_list:
        print(f"{BCOLORS.WARNING}{warning}{BCOLORS.ENDC}", file=sys.stderr)


def print_errors(error_list: list[BoundaryError], source_roots: list[Path]) -> None:
    if not error_list:
        return
    sorted_results = sorted(error_list, key=lambda e: e.file_path)
    for error in sorted_results:
        print(
            build_error_message(error, source_roots=source_roots),
            file=sys.stderr,
        )
    if not all(error.error_info.is_deprecated for error in sorted_results):
        print(
            f"{BCOLORS.WARNING}\nIf you intended to add a new dependency, run 'tach sync' to update your module configuration."
            f"\nOtherwise, remove any disallowed imports and consider refactoring.\n{BCOLORS.ENDC}"
        )


def print_unused_dependencies(
    all_unused_dependencies: list[UnusedDependencies],
) -> None:
    constraint_messages = "\n".join(
        f"\t{BCOLORS.WARNING}'{unused_dependencies.path}' does not depend on: {[dependency.path for dependency in unused_dependencies.dependencies]}{BCOLORS.ENDC}"
        for unused_dependencies in all_unused_dependencies
    )
    print(
        f"❌ {BCOLORS.FAIL}Found unused dependencies: {BCOLORS.ENDC}\n"
        + constraint_messages
    )
    print(
        f"{BCOLORS.WARNING}\nRemove the unused dependencies from tach.yml, "
        f"or consider running 'tach sync' to update module configuration and "
        f"remove all unused dependencies.\n{BCOLORS.ENDC}"
    )


def print_no_config_yml() -> None:
    print(
        f"{BCOLORS.FAIL} {CONFIG_FILE_NAME}.(yml|yaml) not found{BCOLORS.ENDC}",
        file=sys.stderr,
    )


def print_show_web_suggestion() -> None:
    print(
        f"{BCOLORS.OKCYAN}NOTE: You are generating a DOT file locally representing your module graph. For a remotely hosted visualization, use the '--web' argument.\nTo visualize your graph, you will need a program like GraphViz: https://www.graphviz.org/download/\n{BCOLORS.ENDC}"
    )


def print_generated_module_graph_file(output_filepath: Path) -> None:
    print(
        f"{BCOLORS.OKGREEN}Generated a DOT file containing your module graph at '{output_filepath}'{BCOLORS.ENDC}"
    )


def print_circular_dependency_error(cycles: list[list[str]]) -> None:
    print(
        f"❌ {BCOLORS.FAIL}Circular dependencies detected!\n"
        + "\n".join(
            f"{BCOLORS.ENDC}{' -> '.join(cycle)} -> {cycle[0]}" for cycle in cycles
        )
        + f"\n\n{BCOLORS.WARNING}Resolve circular dependencies.\nRemove or unset 'forbid_circular_dependencies' from 'tach.yml' to allow circular dependencies.{BCOLORS.ENDC}"
    )


def add_base_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-e",
        "--exclude",
        required=False,
        type=str,
        metavar="file_or_path,...",
        help="Comma separated path list to exclude. tests/, ci/, etc.",
    )


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="tach",
        add_help=True,
        epilog="Make sure tach is run from the root of your Python project,"
        " and `tach.yml` is present",
    )
    parser.add_argument("--version", action="version", version=f"tach {__version__}")

    subparsers = parser.add_subparsers(title="commands", dest="command")
    mod_parser = subparsers.add_parser(
        "mod",
        prog="tach mod",
        help="Configure module boundaries interactively",
        description="Configure module boundaries interactively",
    )
    mod_parser.add_argument(
        "-d",
        "--depth",
        type=int,
        nargs="?",
        default=None,
        help="The number of child directories to expand from the root",
    )
    add_base_arguments(mod_parser)
    check_parser = subparsers.add_parser(
        "check",
        prog="tach check",
        help="Check existing boundaries against your dependencies and module interfaces",
        description="Check existing boundaries against your dependencies and module interfaces",
    )
    check_parser.add_argument(
        "--exact",
        action="store_true",
        help="Raise errors if any dependency constraints are unused.",
    )
    add_base_arguments(check_parser)
    sync_parser = subparsers.add_parser(
        "sync",
        prog="tach sync",
        help="Sync constraints with actual dependencies in your project.",
        description="Sync constraints with actual dependencies in your project.",
    )
    sync_parser.add_argument(
        "--add",
        action="store_true",
        help="Add any missing dependencies, but do not remove unused dependencies.",
    )
    add_base_arguments(sync_parser)
    report_parser = subparsers.add_parser(
        "report",
        prog="tach report",
        help="Create a report of dependencies and usages of the given path or directory.",
        description="Create a report of dependencies and usages of the given path or directory.",
    )
    report_parser.add_argument(
        "path", help="The path or directory path used to generate the report."
    )
    report_parser.add_argument(
        "-d",
        "--dependencies",
        required=False,
        type=str,
        metavar="module_path,...",
        help="Comma separated module list of dependencies to include [includes everything by default]",
    )
    report_parser.add_argument(
        "--no-deps",
        action="store_true",
        help="Do not include dependencies in the report.",
    )
    report_parser.add_argument(
        "-u",
        "--usages",
        required=False,
        type=str,
        metavar="module_path,...",
        help="Comma separated module list of usages to include [includes everything by default]",
    )
    report_parser.add_argument(
        "--no-usages", action="store_true", help="Do not include usages in the report."
    )
    add_base_arguments(report_parser)
    show_parser = subparsers.add_parser(
        "show",
        prog="tach show",
        help="Visualize the dependency graph of your project.",
        description="Visualize the dependency graph of your project.",
    )
    show_parser.add_argument(
        "--web",
        action="store_true",
        help="Open your dependency graph in a remote web viewer.",
    )
    show_parser.add_argument(
        "-o",
        "--out",
        type=str,
        nargs="?",
        default=None,
        help="Specify an output path for a locally generated module graph file.",
    )
    install_parser = subparsers.add_parser(
        "install",
        prog="tach install",
        help="Install tach into your workflow (e.g. as a pre-commit hook)",
        description="Install tach into your workflow (e.g. as a pre-commit hook)",
    )
    install_parser.add_argument(
        "target",
        choices=InstallTarget.choices(),
        help="What kind of installation to perform (e.g. pre-commit)",
    )
    test_parser = subparsers.add_parser(
        "test",
        prog="tach test",
        help="Run tests on modules impacted by the current changes.",
        description="Run tests on modules impacted by the current changes.",
    )
    test_parser.add_argument(
        "--base",
        type=str,
        nargs="?",
        default="main",
        help="The base commit to use when determining which modules are impacted by changes. [default: 'main']",
    )

    test_parser.add_argument(
        "--head",
        type=str,
        nargs="?",
        default="",
        help="The head commit to use when determining which modules are impacted by changes. [default: current filesystem]",
    )
    test_parser.add_argument(
        "--disable-cache",
        action="store_true",
        help="Do not check cache for results, and do not push results to cache.",
    )
    test_parser.add_argument(
        "pytest_args",
        nargs=argparse.REMAINDER,
        help="Arguments forwarded to pytest. Use '--' to separate these arguments. Ex: 'tach test -- -v'",
    )
    return parser


def parse_arguments(
    args: list[str],
) -> tuple[argparse.Namespace, argparse.ArgumentParser]:
    parser = build_parser()
    parsed_args = parser.parse_args(args)
    return parsed_args, parser


@dataclass
class CachedOutput:
    key: str
    output: list[tuple[int, str]] = field(default_factory=list)
    exit_code: int | None = None

    @property
    def exists(self) -> bool:
        return self.exit_code is not None

    def replay(self):
        for fd, output in self.output:
            if fd == 1:
                print(output, end="", file=sys.stdout)
            elif fd == 2:
                print(output, end="", file=sys.stderr)


def check_cache_for_action(
    project_root: Path, project_config: ProjectConfig, action: str
) -> CachedOutput:
    cache_key = create_computation_cache_key(
        project_root=str(project_root),
        source_roots=[
            str(project_root / source_root)
            for source_root in project_config.source_roots
        ],
        action=action,
        py_interpreter_version=f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        file_dependencies=project_config.cache.file_dependencies,
        env_dependencies=project_config.cache.env_dependencies,
        backend=project_config.cache.backend,
    )
    cache_result = check_computation_cache(
        project_root=str(project_root), cache_key=cache_key
    )
    if cache_result:
        return CachedOutput(
            key=cache_key,
            output=cache_result[0],
            exit_code=cache_result[1],
        )
    return CachedOutput(key=cache_key)


class TeeStream:
    def __init__(self, fd: int, source_stream: IO[Any], capture: list[tuple[int, str]]):
        self.fd = fd
        self.source_stream = source_stream
        self.capture = capture

    def write(self, data: Any):
        self.source_stream.write(data)
        self.capture.append((self.fd, data))

    def __getattr__(self, name: str) -> Any:
        # Hack: Proxy attribute access to the source stream
        return getattr(self.source_stream, name)


class Tee:
    def __init__(self):
        # stdout output will be indicated by (1, <data>), stderr output by (2, <data>)
        self.output_capture: list[tuple[int, str]] = []
        self.original_stdout: Any = None
        self.original_stderr: Any = None

    def __enter__(self):
        self.original_stdout = sys.stdout
        self.original_stderr = sys.stderr

        sys.stdout = TeeStream(1, sys.stdout, self.output_capture)
        sys.stderr = TeeStream(2, sys.stderr, self.output_capture)

        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any):
        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr


def tach_check(
    project_root: Path,
    exact: bool = False,
    exclude_paths: list[str] | None = None,
):
    logger.info(
        "tach check called",
        extra={
            "data": LogDataModel(
                function="tach_check",
                parameters={"exact": exact},
            ),
        },
    )
    try:
        project_config = parse_project_config(project_root)
        if project_config is None:
            print_no_config_yml()
            sys.exit(1)

        exact |= project_config.exact

        check_result = check(
            project_root=project_root,
            project_config=project_config,
            exclude_paths=exclude_paths,
        )
        if check_result.warnings:
            print_warnings(check_result.warnings)

        source_roots = [
            project_root / source_root for source_root in project_config.source_roots
        ]

        if check_result.deprecated_warnings:
            print_errors(
                check_result.deprecated_warnings,
                source_roots=source_roots,
            )
        exit_code = 0

        if check_result.errors:
            print_errors(
                check_result.errors,
                source_roots=source_roots,
            )
            exit_code = 1

        # If we're checking in strict mode, we want to verify that pruning constraints has no effect
        if exact:
            pruned_config = sync_dependency_constraints(
                project_root=project_root,
                project_config=project_config,
                exclude_paths=exclude_paths,
            )
            unused_dependencies = pruned_config.compare_dependencies(project_config)
            if unused_dependencies:
                print_unused_dependencies(unused_dependencies)
                exit_code = 1
    except Exception as e:
        if isinstance(e, TachCircularDependencyError):
            print_circular_dependency_error(e.cycles)
        else:
            print(str(e))
        sys.exit(1)

    if exit_code == 0:
        print(f"✅ {BCOLORS.OKGREEN}All module dependencies validated!{BCOLORS.ENDC}")
    sys.exit(exit_code)


def tach_mod(
    project_root: Path, depth: int | None = 1, exclude_paths: list[str] | None = None
):
    logger.info(
        "tach mod called",
        extra={
            "data": LogDataModel(
                function="tach_mod",
                parameters={"depth": depth},
            ),
        },
    )
    try:
        project_config = parse_project_config(root=project_root) or ProjectConfig()
        saved_changes, warnings = mod_edit_interactive(
            project_root=project_root, project_config=project_config, depth=depth
        )
    except Exception as e:
        print(str(e))
        sys.exit(1)

    if warnings:
        print("\n".join(warnings))
    if saved_changes:
        print(
            f"✅ {BCOLORS.OKGREEN}Set modules! You may want to run '{TOOL_NAME} sync' "
            f"to automatically set boundaries.{BCOLORS.ENDC}"
        )
    sys.exit(0)


def tach_sync(
    project_root: Path, add: bool = False, exclude_paths: list[str] | None = None
):
    logger.info(
        "tach sync called",
        extra={
            "data": LogDataModel(
                function="tach_sync",
                parameters={"add": add},
            ),
        },
    )
    try:
        project_config = parse_project_config(root=project_root)
        if project_config is None:
            print_no_config_yml()
            sys.exit(1)

        if exclude_paths is not None and project_config.exclude is not None:
            exclude_paths.extend(project_config.exclude)
        else:
            exclude_paths = project_config.exclude

        sync_project(
            project_root=project_root,
            project_config=project_config,
            add=add,
            exclude_paths=exclude_paths,
        )
    except Exception as e:
        print(str(e))
        sys.exit(1)

    print(f"✅ {BCOLORS.OKGREEN}Synced dependencies.{BCOLORS.ENDC}")
    sys.exit(0)


class InstallTarget(Enum):
    PRE_COMMIT = "pre-commit"

    @classmethod
    def choices(cls) -> list[str]:
        return [item.value for item in cls]


def tach_install(project_root: Path, target: InstallTarget) -> None:
    logger.info(
        "tach install called",
        extra={
            "data": LogDataModel(
                function="tach_install",
            ),
        },
    )
    try:
        if target == InstallTarget.PRE_COMMIT:
            installed, warning = install_pre_commit(project_root=project_root)
        else:
            raise NotImplementedError(f"Target {target} is not supported by 'install'.")
    except Exception as e:
        print(str(e))
        sys.exit(1)

    if installed:
        print(
            f"✅ {BCOLORS.OKGREEN}Pre-commit hook installed to '.git/hooks/pre-commit'.{BCOLORS.ENDC}"
        )
        sys.exit(0)
    else:
        print(
            f"{BCOLORS.WARNING}Pre-commit hook could not be installed: {warning} {BCOLORS.ENDC}"
        )
        sys.exit(1)


def tach_report(
    project_root: Path,
    path: str,
    include_dependency_modules: list[str] | None = None,
    include_usage_modules: list[str] | None = None,
    skip_dependencies: bool = False,
    skip_usages: bool = False,
    exclude_paths: list[str] | None = None,
):
    logger.info(
        "tach report called",
        extra={
            "data": LogDataModel(
                function="tach_report",
            ),
        },
    )
    project_config = parse_project_config(root=project_root)
    if project_config is None:
        print_no_config_yml()
        sys.exit(1)

    report_path = Path(path)
    try:
        print(
            report(
                project_root,
                report_path,
                project_config=project_config,
                include_dependency_modules=include_dependency_modules,
                include_usage_modules=include_usage_modules,
                skip_dependencies=skip_dependencies,
                skip_usages=skip_usages,
                exclude_paths=exclude_paths,
            )
        )
        sys.exit(0)
    except TachError as e:
        print(f"Report failed: {e}")
        sys.exit(1)


def tach_show(
    project_root: Path, is_web: bool = False, output_filepath: Path | None = None
):
    logger.info(
        "tach show called",
        extra={
            "data": LogDataModel(function="tach_show", parameters={"is_web": is_web}),
        },
    )

    project_config = parse_project_config(root=project_root)
    if project_config is None:
        print_no_config_yml()
        sys.exit(1)

    try:
        if is_web:
            result = generate_show_url(project_config)
            if result:
                print("View your dependency graph here:")
                print(result)
                sys.exit(0)
            else:
                sys.exit(1)
        else:
            print_show_web_suggestion()
            output_filepath = output_filepath or Path("tach_module_graph.dot")
            generate_module_graph_dot_file(project_config, output_filepath)
            print_generated_module_graph_file(output_filepath)
            sys.exit(0)
    except TachError as e:
        print(f"Failed to show module graph: {e}")
        sys.exit(1)


def tach_test(
    project_root: Path,
    head: str,
    base: str,
    disable_cache: bool,
    pytest_args: list[Any],
):
    logger.info(
        "tach test called",
        extra={
            "data": LogDataModel(
                function="tach_test",
            ),
        },
    )
    project_config = parse_project_config(root=project_root)
    if project_config is None:
        print_no_config_yml()
        sys.exit(1)

    if pytest_args and pytest_args[0] != "--":
        print(
            f"{BCOLORS.FAIL}Unknown arguments received. Use '--' to separate arguments for pytest. Ex: 'tach test -- -v'{BCOLORS.ENDC}"
        )
        sys.exit(1)

    try:
        if disable_cache:
            # If cache disabled, just run affected tests and exit
            results = run_affected_tests(
                project_root=project_root,
                project_config=project_config,
                head=head,
                base=base,
                pytest_args=pytest_args[1:],  # Remove '--' pseudo-argument
            )
            sys.exit(results.exit_code)

        # Below this line caching is enabled
        cached_output = check_cache_for_action(
            project_root, project_config, f"tach-test,{head},{base},{pytest_args}"
        )
        if cached_output.exists:
            # Early exit, cached terminal output was found
            print(
                f"{BCOLORS.OKGREEN}============ Cached results found!  ============{BCOLORS.ENDC}"
            )
            cached_output.replay()
            print(
                f"{BCOLORS.OKGREEN}============ END Cached results  ============{BCOLORS.ENDC}"
            )
            sys.exit(cached_output.exit_code)

        # Cache missed, capture terminal output while tests run so we can update the cache

        with Tee() as captured:
            results = run_affected_tests(
                project_root=project_root,
                project_config=project_config,
                head=head,
                base=base,
                pytest_args=pytest_args[1:],  # Remove '--' pseudo-argument
            )

        if results.tests_ran_to_completion:
            update_computation_cache(
                str(project_root),
                cache_key=cached_output.key,
                value=(captured.output_capture, results.exit_code),
            )
        sys.exit(results.exit_code)
    except TachError as e:
        print(f"{BCOLORS.FAIL}Report failed: {e}{BCOLORS.ENDC}")
        sys.exit(1)


def main() -> None:
    args, parser = parse_arguments(sys.argv[1:])
    project_root = fs.find_project_config_root() or Path.cwd()

    latest_version = cache.get_latest_version()
    if latest_version and latest_version != __version__:
        print(
            f"{BCOLORS.WARNING}WARNING: there is a new tach version available"
            f" ({__version__} -> {latest_version}). Upgrade to remove this warning.{BCOLORS.ENDC}"
        )

    # TODO: rename throughout to 'exclude_patterns' to indicate that these are regex patterns
    exclude_paths = args.exclude.split(",") if getattr(args, "exclude", None) else None

    if args.command == "mod":
        tach_mod(
            project_root=project_root, depth=args.depth, exclude_paths=exclude_paths
        )
    elif args.command == "sync":
        tach_sync(project_root=project_root, add=args.add, exclude_paths=exclude_paths)
    elif args.command == "check":
        tach_check(
            project_root=project_root, exact=args.exact, exclude_paths=exclude_paths
        )
    elif args.command == "install":
        try:
            install_target = InstallTarget(args.target)
        except ValueError:
            print(f"{args.target} is not a valid installation target.")
            sys.exit(1)
        tach_install(project_root=project_root, target=install_target)
    elif args.command == "report":
        include_dependency_modules = (
            args.dependencies.split(",") if args.dependencies else None
        )
        include_usage_modules = args.usages.split(",") if args.usages else None
        tach_report(
            project_root=project_root,
            path=args.path,
            include_dependency_modules=include_dependency_modules,
            include_usage_modules=include_usage_modules,
            skip_dependencies=args.no_deps,
            skip_usages=args.no_usages,
            exclude_paths=exclude_paths,
        )
    elif args.command == "test":
        tach_test(
            project_root=project_root,
            head=args.head,
            base=args.base,
            disable_cache=args.disable_cache,
            pytest_args=args.pytest_args,
        )
    elif args.command == "show":
        tach_show(
            project_root=project_root,
            output_filepath=Path(args.out) if args.out is not None else None,
            is_web=args.web,
        )
    else:
        print("Unrecognized command")
        parser.print_help()
        exit(1)


__all__ = ["main"]



from __future__ import annotations

from itertools import chain
from unittest.mock import patch

import pytest

from tach.check import check_import
from tach.cli import tach_check
from tach.core import (
    ModuleConfig,
)
from tach.core.config import RootModuleConfig
from tach.filesystem import validate_project_modules


@pytest.mark.parametrize(
    "valid_modules,invalid_modules",
    [
        (["mod.a"], []),
        ([], ["mod.b"]),
        (["mod.a", "mod.b"], ["mod.c"]),
        (["mod.a", "mod.b"], ["mod.c", "mod.d"]),
    ],
)
def test_validate_project_modules(tmp_path, valid_modules, invalid_modules):
    def mock_fs_check(source_roots, module_path):
        return module_path in valid_modules

    mock_source_root = tmp_path / "src"
    with patch("tach.filesystem.module_to_pyfile_or_dir_path", wraps=mock_fs_check):
        result = validate_project_modules(
            [mock_source_root],
            [ModuleConfig(path=path) for path in chain(valid_modules, invalid_modules)],
        )
        assert set(mod.path for mod in result.valid_modules) == set(valid_modules)
        assert set(mod.path for mod in result.invalid_modules) == set(invalid_modules)


@patch("tach.filesystem.module_to_pyfile_or_dir_path")
def test_validate_project_modules_root_is_always_valid(tmp_path):
    result = validate_project_modules([tmp_path / "src"], [RootModuleConfig()])
    assert (
        len(result.valid_modules) == 1 and result.valid_modules[0] == RootModuleConfig()
    )
    assert not result.invalid_modules


@pytest.mark.parametrize(
    "file_mod_path,import_mod_path,expected_result",
    [
        ("domain_one", "domain_one", True),
        ("domain_one", "domain_one.core", True),
        ("domain_one", "domain_three", True),
        ("domain_two", "domain_one", True),
        ("domain_two", "domain_one.public_fn", True),
        ("domain_two.subdomain", "domain_one", True),
        ("domain_two", "external", True),
        ("external", "external", True),
        ("domain_two", "domain_one.private_fn", False),
        ("domain_three", "domain_one", False),
        ("domain_two", "domain_one.core", False),
        ("domain_two.subdomain", "domain_one.core", False),
        ("domain_two", "domain_three", False),
        ("domain_two", "domain_two.subdomain", False),
        ("external", "domain_three", False),
    ],
)
def test_check_import(module_tree, file_mod_path, import_mod_path, expected_result):
    check_error = check_import(
        module_tree=module_tree,
        file_mod_path=file_mod_path,
        import_mod_path=import_mod_path,
    )
    result = check_error is None
    assert result == expected_result


def test_check_deprecated_import(module_tree):
    check_error = check_import(
        module_tree=module_tree,
        file_mod_path="domain_one",
        import_mod_path="domain_one.subdomain",
    )
    assert check_error is not None
    assert check_error.is_deprecated


def test_valid_example_dir(example_dir, capfd):
    project_root = example_dir / "valid"
    with pytest.raises(SystemExit) as exc_info:
        tach_check(project_root=project_root)
    assert exc_info.value.code == 0
    captured = capfd.readouterr()
    assert "✅" in captured.out  # success state
    assert "‼️" in captured.err  # deprecated warning


def test_valid_example_dir_monorepo(example_dir):
    project_root = example_dir / "monorepo"
    with pytest.raises(SystemExit) as exc_info:
        tach_check(project_root=project_root)
    assert exc_info.value.code == 0



from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
from typing import Generator

from tach.core.config import ModuleConfig, RootModuleConfig


@dataclass
class ModuleNode:
    """
    A node in the module tree.

    If 'is_end_of_path' is True, this node represents a module in the project,
    and must have 'config' and 'full_path' set.

    If 'is_end_of_path' is False, this node does not represent a real module,
    and must have 'config' None and 'full_path' as the empty string.
    """

    is_end_of_path: bool
    full_path: str
    config: ModuleConfig | None
    interface_members: list[str] = field(default_factory=list)
    children: dict[str, ModuleNode] = field(default_factory=dict)

    @classmethod
    def empty(cls) -> ModuleNode:
        return ModuleNode(is_end_of_path=False, full_path="", config=None)

    @classmethod
    def implicit_root(cls) -> ModuleNode:
        config = RootModuleConfig()
        return ModuleNode(is_end_of_path=True, full_path=".", config=config)

    def fill(
        self, config: ModuleConfig, full_path: str, interface_members: list[str]
    ) -> None:
        self.is_end_of_path = True
        self.config = config
        self.full_path = full_path
        self.interface_members = interface_members


def split_module_path(path: str) -> list[str]:
    if not path or path == ".":
        return []
    return path.split(".")


@dataclass
class ModuleTree:
    """
    The core data structure for tach, representing the modules in a project
    with a tree structure for module path lookups.
    """

    root: ModuleNode = field(default_factory=ModuleNode.implicit_root)

    def __iter__(self):
        return module_tree_iterator(self)

    def get(self, path: str) -> ModuleNode | None:
        if not path:
            return None

        node = self.root
        parts = split_module_path(path)

        for part in parts:
            if part not in node.children:
                return None
            node = node.children[part]

        return node if node.is_end_of_path else None

    def insert(self, config: ModuleConfig, path: str, interface_members: list[str]):
        if not path:
            raise ValueError("Cannot insert module with empty path.")

        node = self.root
        parts = split_module_path(path)

        for part in parts:
            if part not in node.children:
                node.children[part] = ModuleNode.empty()
            node = node.children[part]

        node.fill(config, path, interface_members)

    def find_nearest(self, path: str) -> ModuleNode | None:
        node = self.root
        parts = split_module_path(path)
        nearest_parent = node

        for part in parts:
            if part in node.children:
                node = node.children[part]
                if node.is_end_of_path:
                    nearest_parent = node
            else:
                break

        return nearest_parent if nearest_parent.is_end_of_path else None


def module_tree_iterator(tree: ModuleTree) -> Generator[ModuleNode, None, None]:
    stack = deque([tree.root])

    while stack:
        node = stack.popleft()
        if node.is_end_of_path:
            yield node

        stack.extend(node.children.values())



from __future__ import annotations

import logging
import multiprocessing
import os
import signal
from typing import TYPE_CHECKING, Any, Dict

from pydantic import BaseModel, Field

from tach import __version__, cache
from tach.logging.api import log_record, log_uid
from tach.parsing import parse_project_config

if TYPE_CHECKING:
    from types import FrameType


class LogDataModel(BaseModel):
    function: str
    parameters: Dict[str, Any] = Field(default_factory=dict)


def send_log_entry(record: logging.LogRecord, entry: str) -> None:
    is_ci = "CI" in os.environ
    is_gauge = "IS_GAUGE" in os.environ
    data: LogDataModel | None = getattr(record, "data", None)
    uid = cache.get_uid()
    version = __version__
    log_data: dict[str, Any] = {
        "user": str(uid) if uid else None,
        "message": entry,
        "level": record.levelname,
        "timestamp": record.created,
        "function": data.function if data else None,
        "parameters": data.parameters if data else None,
        "version": version,
    }
    if uid is not None:
        log_uid(uid=uid, is_ci=is_ci, is_gauge=is_gauge)
    log_record(record_data=log_data)
    cache.update_latest_version()


def handle_log_entry(record: logging.LogRecord, entry: str) -> None:
    # return on main process
    try:
        if os.fork() != 0:
            return
    except AttributeError:  # TODO WIN support
        return

    import sys

    devnull = open(os.devnull, "w")
    sys.stdout = devnull
    sys.stderr = devnull

    def handler(signum: int, frame: FrameType | None) -> None:
        raise TimeoutError()

    signal.signal(signal.SIGALRM, handler)  # ensure logging process always exits
    signal.alarm(3)  # 3 sec timeout
    try:
        send_log_entry(record=record, entry=entry)
    except Exception:  # noqa
        pass
    finally:
        signal.alarm(0)


class RemoteLoggingHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        log_entry = self.format(record)
        # Ensure logs are nonblocking to main process
        process = multiprocessing.Process(
            target=handle_log_entry, args=(record, log_entry)
        )
        process.start()
        process.join()


logger = logging.getLogger("tach")
logger.setLevel(logging.INFO)
remote_handler = RemoteLoggingHandler()

# Check if logging is enabled
disable_logging = False
project_config = parse_project_config()
if project_config:
    disable_logging = project_config.disable_logging
if not disable_logging:
    logger.addHandler(remote_handler)
else:
    logger.disabled = True



from __future__ import annotations

from typing import TYPE_CHECKING

from tach import errors
from tach.extension import create_dependency_report, set_excluded_paths

if TYPE_CHECKING:
    from pathlib import Path

    from tach.core import ProjectConfig


def report(
    project_root: Path,
    path: Path,
    project_config: ProjectConfig,
    include_dependency_modules: list[str] | None = None,
    include_usage_modules: list[str] | None = None,
    skip_dependencies: bool = False,
    skip_usages: bool = False,
    exclude_paths: list[str] | None = None,
) -> str:
    if not project_root.is_dir():
        raise errors.TachSetupError(
            f"The path '{project_root}' is not a valid directory."
        )

    if not path.exists():
        raise errors.TachError(f"The path '{path}' does not exist.")

    if exclude_paths is not None and project_config.exclude is not None:
        exclude_paths.extend(project_config.exclude)
    else:
        exclude_paths = project_config.exclude

    # This informs the Rust extension ahead-of-time which paths are excluded.
    set_excluded_paths(exclude_paths=exclude_paths or [])

    return create_dependency_report(
        project_root=str(project_root),
        source_roots=[
            str(project_root / source_root)
            for source_root in project_config.source_roots
        ],
        path=str(path),
        include_dependency_modules=include_dependency_modules,
        include_usage_modules=include_usage_modules,
        skip_dependencies=skip_dependencies,
        skip_usages=skip_usages,
        ignore_type_checking_imports=project_config.ignore_type_checking_imports,
    )


__all__ = ["report"]



from __future__ import annotations

import pytest

from tach.cli import tach_sync


def test_valid_example_dir(example_dir, capfd):
    project_root = example_dir / "valid"
    with pytest.raises(SystemExit) as exc_info:
        # TODO not a great test, can change the example fs
        tach_sync(project_root=project_root)
    assert exc_info.value.code == 0
    captured = capfd.readouterr()
    assert "✅" in captured.out  # success state



from __future__ import annotations

from tach.filesystem.install import install_pre_commit
from tach.filesystem.project import find_project_config_root, get_project_config_path
from tach.filesystem.service import (
    ProjectModuleValidationResult,
    canonical,
    chdir,
    delete_file,
    file_to_module_path,
    get_cwd,
    module_to_interface_path,
    module_to_pyfile_or_dir_path,
    parse_ast,
    read_file,
    validate_project_modules,
    walk,
    walk_pyfiles,
    write_file,
)

__all__ = [
    "get_cwd",
    "chdir",
    "canonical",
    "read_file",
    "write_file",
    "delete_file",
    "parse_ast",
    "walk",
    "walk_pyfiles",
    "file_to_module_path",
    "module_to_interface_path",
    "module_to_pyfile_or_dir_path",
    "get_project_config_path",
    "find_project_config_root",
    "install_pre_commit",
    "validate_project_modules",
    "ProjectModuleValidationResult",
]



from __future__ import annotations

import json
from json.decoder import JSONDecodeError
from typing import TYPE_CHECKING
from urllib import error, request

import networkx as nx

if TYPE_CHECKING:
    from pathlib import Path

    import pydot  # type: ignore

    from tach.core import ProjectConfig

TACH_SHOW_URL = "https://show.gauge.sh"


def generate_show_url(project_config: ProjectConfig) -> str | None:
    json_data = project_config.model_dump_json()
    json_bytes = json_data.encode("utf-8")
    req = request.Request(
        f"{TACH_SHOW_URL}/api/core/0.6.5/graph/",
        data=json_bytes,
        headers={"Content-Type": "application/json"},
    )

    try:
        # Send the request and read the response
        with request.urlopen(req) as response:
            response_data = response.read().decode("utf-8")
            response_json = json.loads(response_data)
            url = response_json.get("uid")
            return f"{TACH_SHOW_URL}?uid={url}"
    except (UnicodeDecodeError, JSONDecodeError, error.URLError) as e:
        print(f"Error: {e}")
        return None


def generate_module_graph_dot_file(
    project_config: ProjectConfig, output_filepath: Path
) -> None:
    graph = nx.DiGraph()  # type: ignore
    # Add nodes
    for module in project_config.modules:
        graph.add_node(module.path)  # type: ignore

    # Add dependency edges
    for module in project_config.modules:
        for dependency in module.depends_on:
            graph.add_edge(module.path, dependency)  # type: ignore

    pydot_graph: pydot.Dot = nx.nx_pydot.to_pydot(graph)  # type: ignore
    dot_data: str = pydot_graph.to_string()  # type: ignore

    output_filepath.write_text(dot_data)  # type: ignore


__all__ = ["generate_show_url", "generate_module_graph_dot_file"]



from __future__ import annotations


class TachError(Exception): ...


class TachParseError(TachError): ...


class TachSetupError(TachError): ...


class TachCircularDependencyError(TachError):
    def __init__(self, cycles: list[list[str]]):
        self.cycles = cycles


__all__ = [
    "TachError",
    "TachParseError",
    "TachSetupError",
    "TachCircularDependencyError",
]



from __future__ import annotations

from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

from pydantic import AfterValidator, BaseModel, Field, field_serializer
from typing_extensions import Annotated, Literal

from tach.constants import DEFAULT_EXCLUDE_PATHS, ROOT_MODULE_SENTINEL_TAG


class Config(BaseModel):
    model_config = {"extra": "forbid"}


class ModuleConfig(Config):
    """
    Configuration for a single module in a Tach project.

    Primarily responsible for declaring dependencies.
    """

    path: str
    depends_on: List[Dependency] = Field(default_factory=list)
    strict: bool = False

    @property
    def mod_path(self) -> str:
        if self.path == ROOT_MODULE_SENTINEL_TAG:
            return "."
        return self.path


class Dependency(Config):
    path: str
    deprecated: bool = False

    model_config = {"frozen": True}


def validate_root_path(path: str) -> str:
    assert path == ROOT_MODULE_SENTINEL_TAG
    return path


class RootModuleConfig(ModuleConfig):
    """
    Special-case schema for the implicit root module configuration.
    """

    path: Annotated[str, AfterValidator(validate_root_path)] = ROOT_MODULE_SENTINEL_TAG


@dataclass
class UnusedDependencies:
    path: str
    dependencies: List[Dependency]


class CacheConfig(Config):
    """
    Configuration affecting Tach's caching.

    Responsible for configuring the cache backend, and adjusting the dependencies which should affect cache retrieval.
    """

    backend: Literal["local"] = "local"
    file_dependencies: List[str] = Field(default_factory=list)
    env_dependencies: List[str] = Field(default_factory=list)


class ProjectConfig(Config):
    """
    Central configuration object for a project using Tach.

    Controls which modules are defined, their dependencies, as well as global tool-related configuration.
    """

    modules: List[ModuleConfig] = Field(default_factory=list)
    cache: CacheConfig = Field(default_factory=CacheConfig)
    exclude: Optional[List[str]] = Field(
        default_factory=lambda: copy(DEFAULT_EXCLUDE_PATHS)
    )
    source_roots: List[Path] = Field(default_factory=lambda: [Path(".")])
    exact: bool = False
    disable_logging: bool = False
    ignore_type_checking_imports: bool = True
    forbid_circular_dependencies: bool = False

    @field_serializer("source_roots")
    def serialize_source_roots(self, source_roots: list[Path], _) -> List[str]:
        return list(map(str, source_roots))

    @property
    def module_paths(self) -> list[str]:
        return [module.path for module in self.modules]

    def set_modules(self, module_paths: list[str]) -> None:
        new_modules: list[ModuleConfig] = []
        new_module_paths = set(module_paths)
        original_modules_by_path = {
            module_config.path: module_config for module_config in self.modules
        }
        for new_module_path in new_module_paths:
            if new_module_path in original_modules_by_path:
                original_module = original_modules_by_path[new_module_path]
                original_module.depends_on = [
                    dep
                    for dep in original_module.depends_on
                    if dep.path in new_module_paths
                ]
                new_modules.append(original_module)
            else:
                new_modules.append(ModuleConfig(path=new_module_path))

        self.modules = new_modules

    def dependencies_for_module(self, module: str) -> list[Dependency]:
        return next(
            (mod.depends_on for mod in self.modules if mod.path == module),
            list(),  # type: ignore
        )

    def add_dependency_to_module(self, module: str, dependency: Dependency):
        current_module_config = next(
            (
                module_config
                for module_config in self.modules
                if module_config.path == module
            ),
            None,
        )
        if not current_module_config:
            # No configuration exists for tag, add default config with this dependency
            self.modules.append(ModuleConfig(path=module, depends_on=[dependency]))
        else:
            # Config already exists, set the union of existing dependencies and new ones
            new_dependencies = set(current_module_config.depends_on) | {dependency}  # pyright: ignore[reportUnhashable]
            current_module_config.depends_on = list(new_dependencies)

    def compare_dependencies(
        self, other_config: ProjectConfig
    ) -> list[UnusedDependencies]:
        all_unused_dependencies: list[UnusedDependencies] = []
        own_module_paths = set(module.path for module in self.modules)
        for module_config in other_config.modules:
            if module_config.path not in own_module_paths:
                all_unused_dependencies.append(
                    UnusedDependencies(
                        path=module_config.path, dependencies=module_config.depends_on
                    )
                )
                continue
            own_module_dependency_paths = set(
                dep.path
                for dep in self.dependencies_for_module(module=module_config.path)
            )
            current_dependency_paths = set(dep.path for dep in module_config.depends_on)
            extra_dependency_paths = (
                current_dependency_paths - own_module_dependency_paths
            )
            if extra_dependency_paths:
                extra_dependencies = [
                    dep
                    for dep in module_config.depends_on
                    if dep.path in extra_dependency_paths
                ]
                all_unused_dependencies.append(
                    UnusedDependencies(
                        path=module_config.path,
                        dependencies=extra_dependencies,
                    )
                )

        return all_unused_dependencies



from __future__ import annotations

from pathlib import Path
from typing import Any

import pydantic
import yaml

from tach import filesystem as fs
from tach.constants import ROOT_MODULE_SENTINEL_TAG, TACH_YML_SCHEMA_URL
from tach.core import ProjectConfig


class TachYamlDumper(yaml.Dumper):
    def increase_indent(self, flow: bool = False, indentless: bool = False):
        return super().increase_indent(flow, False)


def dump_project_config_to_yaml(config: ProjectConfig) -> str:
    # Using sort_keys=False here and depending on config.model_dump maintaining 'insertion order'
    # so that 'tag' appears before 'depends_on'
    # Instead, should provide custom yaml.Dumper & yaml.Representer or just write our own
    # Sort only constraints and dependencies alphabetically for now
    config.modules.sort(
        key=lambda mod: (mod.path == ROOT_MODULE_SENTINEL_TAG, mod.path)
    )
    for mod in config.modules:
        mod.depends_on.sort(key=lambda dep: dep.path)
    # NOTE: setting 'exclude' explicitly here also interacts with the 'exclude_unset' option
    # being passed to 'model_dump'. It ensures that even on a fresh config, we will explicitly
    # show excluded paths.
    config.exclude = list(set(config.exclude)) if config.exclude else []
    config.exclude.sort()
    language_server_directive = (
        f"# yaml-language-server: $schema={TACH_YML_SCHEMA_URL}\n"
    )
    yaml_content = yaml.dump(
        config.model_dump(exclude_unset=True),
        Dumper=TachYamlDumper,
        sort_keys=False,
        default_flow_style=False,
        indent=2,
    )
    return language_server_directive + yaml_content


# TODO remove after next major version upgrade
def migrate_config(result: dict[Any, Any]) -> dict[Any, Any]:
    if "modules" in result:
        for module in result["modules"]:
            if "depends_on" in module:
                for index, path in enumerate(module["depends_on"]):
                    if isinstance(path, str):
                        module["depends_on"][index] = {"path": path}
    if "source_root" in result and isinstance(result["source_root"], str):
        result["source_roots"] = [result["source_root"]]
        del result["source_root"]
    return result


def parse_project_config(root: Path | None = None) -> ProjectConfig | None:
    root = root or Path.cwd()
    file_path = fs.get_project_config_path(root)
    if not file_path:
        return None

    with open(file_path) as f:
        result = yaml.safe_load(f)
        if not result or not isinstance(result, dict):
            raise ValueError(f"Empty or invalid project config file: {file_path}")
    try:
        config = ProjectConfig(**result)  # type: ignore
    except pydantic.ValidationError:
        result = migrate_config(result)  # type: ignore
        config = ProjectConfig(**result)
        print("Updating config to latest syntax...")
        config_yml_content = dump_project_config_to_yaml(config)
        fs.write_file(str(file_path), config_yml_content)
    return config



from __future__ import annotations

import ast
from typing import Any


class EarlyExitNodeVisitor(ast.NodeVisitor):
    def __init__(self, *args: list[Any], **kwargs: dict[Any, Any]):
        self._exit = False

    def set_exit(self, flag: bool = True):
        self._exit = flag

    def visit(self, node: ast.AST):
        if self._exit:
            # Allow reusing the instance
            self._exit = False
            return
        super().visit(node)



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING

from tach import errors
from tach import filesystem as fs
from tach.colors import BCOLORS
from tach.constants import CONFIG_FILE_NAME
from tach.interactive import (
    InteractiveModuleConfiguration,
    get_selected_modules_interactive,
)
from tach.parsing import dump_project_config_to_yaml

if TYPE_CHECKING:
    from tach.core import ProjectConfig


def update_modules(
    project_config: ProjectConfig,
    project_root: Path,
    selected_source_roots: list[Path],
    selected_modules: list[Path],
):
    if set(project_config.source_roots) != set(selected_source_roots):
        # Only assign to this field if it has changed,
        # since the project config writes any field that
        # has been touched out to YML.
        project_config.source_roots = [
            source_root.relative_to(project_root)
            for source_root in selected_source_roots
        ]

    module_paths = [
        fs.file_to_module_path(
            source_roots=tuple(selected_source_roots),
            file_path=selected_module_file_path,
        )
        for selected_module_file_path in selected_modules
    ]
    project_config.set_modules(module_paths=module_paths)

    project_config_path = project_root / f"{CONFIG_FILE_NAME}.yml"
    config_yml_content = dump_project_config_to_yaml(project_config)
    fs.write_file(str(project_config_path), config_yml_content)


@dataclass
class ValidationResult:
    ok: bool
    errors: list[str] = field(default_factory=list)


def validate_configuration(
    configuration: InteractiveModuleConfiguration,
) -> ValidationResult:
    errors: list[str] = []
    for module_path in configuration.module_paths:
        module_path = Path(module_path).resolve()

        if not any(
            source_root in module_path.parents
            for source_root in configuration.source_roots
        ):
            # This module exists outside of the source root
            # This is not allowed and should be reported as a configuration error
            errors.append(
                f"Module '{module_path}' is not contained within any source root: {[str(root) for root in configuration.source_roots]}"
            )
    return ValidationResult(ok=not errors, errors=errors)


def mod_edit_interactive(
    project_root: Path, project_config: ProjectConfig, depth: int | None = 1
) -> tuple[bool, list[str]]:
    if not Path(project_root).is_dir():
        raise errors.TachSetupError(f"The path {project_root} is not a directory.")

    interactive_module_configuration = get_selected_modules_interactive(
        path=project_root,
        project_config=project_config,
        depth=depth,
    )
    if interactive_module_configuration is not None:
        validation_result = validate_configuration(interactive_module_configuration)
        if validation_result.errors:
            return False, [
                f"{BCOLORS.FAIL}Validation error: {BCOLORS.WARNING}{error}{BCOLORS.ENDC}"
                for error in validation_result.errors
            ]
        update_modules(
            project_config=project_config,
            project_root=project_root,
            selected_source_roots=interactive_module_configuration.source_roots,
            selected_modules=interactive_module_configuration.module_paths,
        )
        return True, []
    else:
        return False, [f"{BCOLORS.OKCYAN}No changes saved.{BCOLORS.ENDC}"]


__all__ = ["mod_edit_interactive"]



from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Any

from tach import filesystem as fs
from tach.parsing.ast_visitor import EarlyExitNodeVisitor

if TYPE_CHECKING:
    from pathlib import Path


class InterfaceVisitor(EarlyExitNodeVisitor):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.members: list[str] = []

    def visit_Assign(self, node: ast.Assign):
        # Check if the node is an assignment
        if any(
            isinstance(target, ast.Name) and target.id == "__all__"
            for target in node.targets
        ):
            # Ensure the value is a list and process each element
            if isinstance(node.value, ast.List):
                for element in node.value.elts:
                    if isinstance(element, ast.Str):  # pyright: ignore
                        # Supports Python 3.7 and below
                        self.members.append(element.s)
                    elif isinstance(element, ast.Constant):
                        # Supports Python 3.8 and above
                        if isinstance(element.value, str):
                            self.members.append(element.value)
                # Early exit
                self.set_exit(True)
        # Continue to the next node
        self.generic_visit(node)


def parse_interface_members(source_roots: list[Path], module_path: str) -> list[str]:
    """
    Parse the members of __all__ in a given module
    """
    file_path = fs.module_to_interface_path(tuple(source_roots), module_path)
    if file_path is None:
        return []

    parsed_ast = fs.parse_ast(str(file_path))
    interface_visitor = InterfaceVisitor()
    interface_visitor.visit(parsed_ast)
    return interface_visitor.members
